from datetime import datetime
from itertools import chain
from typing import Any, Awaitable, Callable, NamedTuple
from uuid import UUID

import structlog
from more_itertools import flatten, unique_everseen
from sqlalchemy import Row, func, select
from sqlalchemy import Text as SaText
from sqlalchemy import cast as sa_cast
from sqlalchemy.orm import aliased

from orchestrator.db import (
    ProcessSubscriptionTable,
    ProcessTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import (
    SubscriptionInstanceRelationTable,
)
from orchestrator.services.subscriptions import RELATION_RESOURCE_TYPES
from orchestrator.types import SubscriptionLifecycle

logger = structlog.get_logger(__name__)


class Relation(NamedTuple):
    depends_on_sub_id: UUID
    in_use_by_sub_id: UUID


def _get_instance_relations(instance_relations_query: Any) -> list[Relation]:
    def to_relation(row: Row[Any]) -> Relation:
        return Relation(row[0], row[1])

    return [to_relation(row) for row in db.session.execute(instance_relations_query)]


async def _get_in_use_by_instance_relations(
    subscription_ids: list[UUID], filter_statuses: tuple[str, ...]
) -> list[Relation]:
    """Get in_use_by by relations through subscription instance hierarchy."""
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    depends_on_instances = aliased(SubscriptionInstanceTable)

    query_get_in_use_by_ids = (
        select(depends_on_instances.subscription_id, in_use_by_instances.subscription_id)
        .distinct()
        .join(in_use_by_instances.subscription)
        .join(in_use_by_instances.depends_on_block_relations)
        .join(depends_on_instances, SubscriptionInstanceRelationTable.depends_on)
        .filter(depends_on_instances.subscription_id.in_(set(subscription_ids)))
        .filter(in_use_by_instances.subscription_id != depends_on_instances.subscription_id)
        .filter(SubscriptionTable.status.in_(filter_statuses))
    )

    return _get_instance_relations(query_get_in_use_by_ids)


async def _get_depends_on_instance_relations(
    subscription_ids: list[UUID], filter_statuses: tuple[str, ...]
) -> list[Relation]:
    """Get depends_on relations through subscription instance hierarchy."""
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    depends_on_instances = aliased(SubscriptionInstanceTable)

    query_get_depends_on_ids = (
        select(depends_on_instances.subscription_id, in_use_by_instances.subscription_id)
        .distinct()
        .join(depends_on_instances.subscription)
        .join(depends_on_instances.in_use_by_block_relations)
        .join(in_use_by_instances, SubscriptionInstanceRelationTable.in_use_by)
        .filter(in_use_by_instances.subscription_id.in_(set(subscription_ids)))
        .filter(depends_on_instances.subscription_id != in_use_by_instances.subscription_id)
        .filter(SubscriptionTable.status.in_(filter_statuses))
    )

    return _get_instance_relations(query_get_depends_on_ids)


def _get_resource_type_relations(resource_type_relations_query: Any) -> list[Relation]:
    def to_relation(row: Row[Any]) -> Relation:
        return Relation(UUID(row[0]), row[1])

    return [to_relation(row) for row in db.session.execute(resource_type_relations_query)]


async def _get_in_use_by_resource_type_relations(
    subscription_ids: list[UUID], filter_statuses: tuple[str, ...]
) -> list[Relation]:
    """Get in_use_by relations through resource types."""
    logger.warning("Using legacy RELATION_RESOURCE_TYPES to find in_use_by subs")

    in_use_by_subscriptions = aliased(SubscriptionTable)
    depends_on_instance_values = aliased(SubscriptionInstanceValueTable)

    # Convert UUIDs to string
    unique_subscription_ids = set(map(str, subscription_ids))

    query_get_in_use_by_ids = (
        select(depends_on_instance_values.value, in_use_by_subscriptions.subscription_id)
        .select_from(depends_on_instance_values)
        .join(SubscriptionInstanceTable)
        .join(in_use_by_subscriptions)
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(depends_on_instance_values.value.in_(unique_subscription_ids))
        .filter(in_use_by_subscriptions.status.in_(filter_statuses))
    )

    return _get_resource_type_relations(query_get_in_use_by_ids)


async def _get_depends_on_resource_type_relations(
    subscription_ids: list[UUID], filter_statuses: tuple[str, ...]
) -> list[Relation]:
    """Get depends_on relations through resource types."""
    logger.warning("Using legacy RELATION_RESOURCE_TYPES to find depends_on subs")

    depends_on_subscriptions = aliased(SubscriptionTable)
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    in_use_by_instance_values = aliased(SubscriptionInstanceValueTable)

    unique_subscription_ids = set(subscription_ids)

    query_get_depends_on_ids = (
        select(in_use_by_instance_values.value, in_use_by_instances.subscription_id)
        .select_from(in_use_by_instance_values)
        .join(in_use_by_instances)
        .join(
            depends_on_subscriptions,
            in_use_by_instance_values.value == sa_cast(depends_on_subscriptions.subscription_id, SaText),
        )
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(in_use_by_instances.subscription_id.in_(unique_subscription_ids))
        .filter(depends_on_subscriptions.status.in_(filter_statuses))
    )

    return _get_resource_type_relations(query_get_depends_on_ids)


async def _get_in_use_by_relations(subscription_ids: list[UUID], filter_statuses: tuple[str, ...]) -> list[Relation]:
    if RELATION_RESOURCE_TYPES:
        # Find relations through resource types
        resource_type_relations = await _get_in_use_by_resource_type_relations(subscription_ids, filter_statuses)
    else:
        resource_type_relations = []
    # Find relations through instance hierarchy
    instance_relations = await _get_in_use_by_instance_relations(subscription_ids, filter_statuses)
    return list(chain(resource_type_relations, instance_relations))


async def _get_depends_on_relations(subscription_ids: list[UUID], filter_statuses: tuple[str, ...]) -> list[Relation]:
    if RELATION_RESOURCE_TYPES:
        # Find relations through resource types
        resource_type_relations = await _get_depends_on_resource_type_relations(subscription_ids, filter_statuses)
    else:
        resource_type_relations = []
    # Find relations through instance hierarchy
    instance_relations = await _get_depends_on_instance_relations(subscription_ids, filter_statuses)
    return list(chain(resource_type_relations, instance_relations))


async def get_in_use_by_subscriptions(
    subscription_ids: list[UUID], filter_statuses: tuple[str, ...]
) -> list[list[SubscriptionTable]]:
    """Function to efficiently get the in_use_by SubscriptionTables for multiple subscription_ids."""
    _filter_statuses: tuple[str, ...] = filter_statuses or tuple(SubscriptionLifecycle.values())
    in_use_by_relations = await _get_in_use_by_relations(subscription_ids, _filter_statuses)

    # Retrieve SubscriptionTable for all unique inuseby ids
    unique_in_use_by_ids = {row.in_use_by_sub_id for row in in_use_by_relations}
    _in_use_by_subs = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id.in_(unique_in_use_by_ids))
    ).scalars()
    in_use_by_subs = {subscription.subscription_id: subscription for subscription in _in_use_by_subs}

    # group (more_itertools.bucket doesn't seem to work for tuple of uuids)
    subscription_in_use_by_ids: dict[UUID, list[UUID]] = {}
    for relation in in_use_by_relations:
        subscription_in_use_by_ids.setdefault(relation.depends_on_sub_id, []).append(relation.in_use_by_sub_id)

    def get_in_use_by_subs(depends_on_id: UUID) -> list[SubscriptionTable]:
        in_use_by_ids = subscription_in_use_by_ids.get(depends_on_id, [])
        return [in_use_by_sub for id_ in in_use_by_ids if (in_use_by_sub := in_use_by_subs.get(id_))]

    # Important (as with any dataloader)
    # Return the list of inuseby subs in the exact same order as the ids passed to this function
    return [get_in_use_by_subs(subscription_id) for subscription_id in subscription_ids]


async def get_depends_on_subscriptions(
    subscription_ids: list[UUID], filter_statuses: tuple[str, ...]
) -> list[list[SubscriptionTable]]:
    """Function to efficiently get the depends_on SubscriptionTables for multiple subscription_ids."""
    _filter_statuses: tuple[str, ...] = filter_statuses or tuple(SubscriptionLifecycle.values())
    depends_on_relations = await _get_depends_on_relations(subscription_ids, _filter_statuses)

    # Retrieve SubscriptionTable for all unique dependson ids
    unique_depends_on_ids = {row.depends_on_sub_id for row in depends_on_relations}
    _depends_on_subs = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id.in_(unique_depends_on_ids))
    ).scalars()
    depends_on_subs = {subscription.subscription_id: subscription for subscription in _depends_on_subs}

    # group (more_itertools.bucket doesn't seem to work for tuple of uuids)
    subscription_depends_on_ids: dict[UUID, list[UUID]] = {}
    for relation in depends_on_relations:
        subscription_depends_on_ids.setdefault(relation.in_use_by_sub_id, []).append(relation.depends_on_sub_id)

    def get_depends_on_subs(in_use_by_id: UUID) -> list[SubscriptionTable]:
        depends_on_ids = subscription_depends_on_ids.get(in_use_by_id, [])
        return [depends_on_sub for id_ in depends_on_ids if (depends_on_sub := depends_on_subs.get(id_))]

    # Important (as with any dataloader)
    # Return the list of dependson subs in the exact same order as the ids passed to this function
    return [get_depends_on_subs(subscription_id) for subscription_id in subscription_ids]


async def get_recursive_relations(
    subscription_ids: list[UUID],
    filter_statuses: tuple[str, ...],
    recurse_product_types: list[str],
    recurse_depth_limit: int,
    relation_fetcher: Callable[[list[UUID], tuple[str, ...]], Awaitable[list[list[SubscriptionTable]]]],
    current_depth: int = 0,
    used_subscription_ids: set[UUID] | None = None,
) -> list[SubscriptionTable]:
    """Recursively fetches subscription relations based on a custom relation-fetching function.

    Args:
        subscription_ids: List of subscription IDs to start with.
        filter_statuses: List of statuses to filter the relations.
        recurse_product_types: List of product types to recurse into.
        recurse_depth_limit: Maximum recursion depth.
        current_depth: Current recursion depth.
        used_subscription_ids: List of already used subscription IDs to avoid cycles.
        relation_fetcher: A function that fetches the relations for the subscriptions.

    Returns:
        A flattened list of all fetched relations for the given subscriptions.
    """

    used_subscription_ids = used_subscription_ids or set(subscription_ids)
    _relations = await relation_fetcher(subscription_ids, filter_statuses)
    relations = list(unique_everseen(flatten(_relations), key=lambda s: s.subscription_id))

    def get_related_subscription_ids() -> list[UUID]:
        return [
            r.subscription_id
            for r in relations
            if r.product.product_type in recurse_product_types and r.subscription_id not in used_subscription_ids
        ]

    nested_relations: list[SubscriptionTable] = []
    if recurse_depth_limit > current_depth and (related_subscription_ids := get_related_subscription_ids()):
        used_subscription_ids.update(related_subscription_ids)
        nested_relations = await get_recursive_relations(
            related_subscription_ids,
            filter_statuses,
            recurse_product_types,
            recurse_depth_limit,
            current_depth=current_depth + 1,
            used_subscription_ids=used_subscription_ids,
            relation_fetcher=relation_fetcher,
        )
    return list(unique_everseen(relations + nested_relations, key=lambda s: s.subscription_id))


async def get_last_validation_datetimes(subscription_ids: list[UUID]) -> list[datetime | None]:
    stmt = (
        select(ProcessSubscriptionTable.subscription_id, func.max(ProcessTable.last_modified_at))
        .join(ProcessSubscriptionTable)
        .group_by(ProcessSubscriptionTable.subscription_id)
        .where(
            (ProcessSubscriptionTable.workflow_target == "VALIDATE")
            & ProcessSubscriptionTable.subscription_id.in_(subscription_ids)
        )
    )
    results = db.session.execute(stmt).all()
    last_validation_indexed_by_sub_id = {
        str(subscription_id): last_validation for subscription_id, last_validation in results
    }
    return [last_validation_indexed_by_sub_id.get(str(subscription_id), None) for subscription_id in subscription_ids]



"""Helper script to rewrite import statements in your orchestrator.

Since types have been externalised in `pydantic_forms`, they were re-imported in `orchestrator.types` for backwards
compatibility. These import statements have been removed, and therefore need to be updated in orchestrator
implementations.
"""

import sys
from pathlib import Path

from orchestrator.devtools.scripts.shared import migrate, move_import


def migrate_file(f: Path) -> bool:
    imports = {
        "JSON": move_import(f, "JSON", "orchestrator.types", "pydantic_forms.types"),
        "AcceptData": move_import(f, "AcceptData", "orchestrator.types", "pydantic_forms.types"),
        "AcceptItemType": move_import(f, "AcceptItemType", "orchestrator.types", "pydantic_forms.types"),
        "FormGenerator": move_import(f, "FormGenerator", "orchestrator.types", "pydantic_forms.types"),
        "FormGeneratorAsync": move_import(f, "FormGeneratorAsync", "orchestrator.types", "pydantic_forms.types"),
        "InputForm": move_import(f, "InputForm", "orchestrator.types", "pydantic_forms.types"),
        "InputFormGenerator": move_import(f, "InputFormGenerator", "orchestrator.types", "pydantic_forms.types"),
        "InputStepFunc": move_import(f, "InputStepFunc", "orchestrator.types", "pydantic_forms.types"),
        "SimpleInputFormGenerator": move_import(
            f, "SimpleInputFormGenerator", "orchestrator.types", "pydantic_forms.types"
        ),
        "State": move_import(f, "State", "orchestrator.types", "pydantic_forms.types"),
        "StateInputFormGenerator": move_import(
            f, "StateInputFormGenerator", "orchestrator.types", "pydantic_forms.types"
        ),
        "StateInputFormGeneratorAsync": move_import(
            f, "StateInputFormGeneratorAsync", "orchestrator.types", "pydantic_forms.types"
        ),
        "StateInputStepFunc": move_import(f, "StateInputStepFunc", "orchestrator.types", "pydantic_forms.types"),
        "StateSimpleInputFormGenerator": move_import(
            f, "StateSimpleInputFormGenerator", "orchestrator.types", "pydantic_forms.types"
        ),
        "SubscriptionMapping": move_import(f, "SubscriptionMapping", "orchestrator.types", "pydantic_forms.types"),
        "SummaryData": move_import(f, "SummaryData", "orchestrator.types", "pydantic_forms.types"),
        "UUIDstr": move_import(f, "UUIDstr", "orchestrator.types", "pydantic_forms.types"),
        "strEnum": move_import(f, "strEnum", "orchestrator.types", "pydantic_forms.types"),
    }
    lines = []
    lines.extend([f"Moved {k} import" for k, v in imports.items() if v])

    if lines:
        formatted_lines = "\n".join(f" - {line}" for line in lines)
        print(f"Updated {f.name:50s}\n{formatted_lines}")

    return bool(lines)


if __name__ == "__main__":
    try:
        _target_dir = Path(sys.argv[1])
        assert _target_dir.is_dir()
    except Exception:
        print("Need a directory as parameter")
        sys.exit(1)

    sys.exit(0 if migrate(_target_dir, migrate_file) else 1)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AG-UI adapter — orchestration and custom transport layer.

AGUIWorker owns the full AG-UI request lifecycle (state setup, persistence,
streaming) so the HTTP endpoint stays a thin request/response handler.

AGUIEventStream / _AGUIAdapter customise the pydantic-ai transport:
1. Replace full tool results with lightweight ToolArtifact references
2. Pass through CustomEvent instances (e.g. AGENT_STEP_ACTIVE) that
   pydantic-ai's default handle_event() would silently drop
"""

from __future__ import annotations

from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import Any
from uuid import UUID

from ag_ui.core import BaseEvent, CustomEvent, EventType, RunAgentInput, ToolCallResultEvent
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.messages import FunctionToolResultEvent, ToolReturnPart
from pydantic_ai.ui import NativeEvent
from pydantic_ai.ui.ag_ui import AGUIAdapter
from pydantic_ai.ui.ag_ui import AGUIEventStream as _BaseAGUIEventStream
from sqlalchemy.orm import Session
from structlog import get_logger

from orchestrator.db.models import AgentRunTable
from orchestrator.search.agent.artifacts import ToolArtifact
from orchestrator.search.agent.persistence import PostgresStatePersistence
from orchestrator.search.agent.state import SearchState

if __name__ != "__main__":
    from orchestrator.search.agent.agent import AgentAdapter

logger = get_logger(__name__)


@dataclass
class AGUIEventStream(_BaseAGUIEventStream[Any, Any]):
    """Custom event stream for the search agent.

    - Replaces tool results that carry ToolArtifact metadata with the artifact JSON,
      so the AG-UI frontend receives a lightweight reference instead of full data.
    - Yields CustomEvent instances (AGENT_STEP_ACTIVE) that the base class would drop.
    """

    async def handle_event(self, event: NativeEvent | CustomEvent) -> AsyncIterator[BaseEvent]:
        # Pass through AG-UI CustomEvents (e.g. AGENT_STEP_ACTIVE) that the
        # base class match/case would silently discard.
        # NativeEvent doesn't include CustomEvent in its union, but we receive
        # them at runtime via the agent's event stream.
        if isinstance(event, CustomEvent):
            yield event
            return

        async for e in super().handle_event(event):
            yield e

    async def handle_function_tool_result(self, event: FunctionToolResultEvent) -> AsyncIterator[BaseEvent]:
        result = event.result
        if isinstance(result, ToolReturnPart) and isinstance(result.metadata, ToolArtifact):
            yield ToolCallResultEvent(
                message_id=self.new_message_id(),
                type=EventType.TOOL_CALL_RESULT,
                role="tool",
                tool_call_id=result.tool_call_id,
                content=result.metadata.model_dump_json(),
            )
            return

        # Default behavior for all other tools
        async for e in super().handle_function_tool_result(event):
            yield e


class _AGUIAdapter(AGUIAdapter[Any, Any]):
    """AGUIAdapter that uses AGUIEventStream."""

    def build_event_stream(self) -> AGUIEventStream:
        return AGUIEventStream(self.run_input, accept=self.accept)


class AGUIWorker:
    """Orchestrates AG-UI request handling: state setup, persistence, stream creation.

    Parallel to A2AWorker — owns the full AG-UI lifecycle so the HTTP endpoint
    stays a thin request/response handler.
    """

    @staticmethod
    async def run_request(
        agent: AgentAdapter,
        run_input: RunAgentInput,
        db_session: Session,
    ) -> AsyncIterator[str]:
        """Execute the full AG-UI lifecycle and return an SSE event iterator.

        Steps:
        1. Extract user input from messages and inject into state
        2. Create/fetch an AgentRunTable record
        3. Set up persistence and load previous state if available
        4. Create the _AGUIAdapter, run the stream, and yield SSE events
        5. Commit on success, rollback on failure
        """
        prepared = AGUIWorker._prepare_run_input(run_input)

        run_id = UUID(prepared.run_id)
        thread_id = prepared.thread_id

        # Create or get agent run record
        agent_run = db_session.get(AgentRunTable, run_id)
        if not agent_run:
            agent_run = AgentRunTable(run_id=run_id, thread_id=thread_id, agent_type="search")
            db_session.add(agent_run)
            db_session.commit()
            logger.debug("Created new agent run", run_id=str(run_id), thread_id=thread_id)

        prepared.state["run_id"] = run_id

        persistence = PostgresStatePersistence(thread_id=thread_id, run_id=run_id, session=db_session)

        loaded_state = await persistence.load_state()
        if loaded_state:
            initial_state = loaded_state
            initial_state.user_input = prepared.state["user_input"]
            initial_state.run_id = run_id
            logger.debug(
                "Loaded previous state from persistence",
                completed_turns=len(initial_state.memory.completed_turns),
                current_turn=initial_state.memory.current_turn is not None,
            )
        else:
            initial_state = SearchState(**prepared.state)
            logger.debug("Created fresh state (no previous snapshot)")

        # Set persistence on agent instance
        agent._persistence = persistence

        # Create adapter — all event handling (artifacts, custom events) flows through the stream
        adapter = _AGUIAdapter(agent=agent, run_input=prepared)
        event_stream = adapter.encode_stream(adapter.run_stream(deps=StateDeps(initial_state)))

        async def _stream_with_persistence() -> AsyncIterator[str]:
            """Stream AG-UI events and commit DB on completion."""
            try:
                async for event_str in event_stream:
                    yield event_str
                db_session.commit()
            except Exception as e:
                logger.error("Error in agent stream", error=str(e), exc_info=True)
                db_session.rollback()
                raise

        return _stream_with_persistence()

    @staticmethod
    def _prepare_run_input(run_input: RunAgentInput) -> RunAgentInput:
        """Prepare RunAgentInput by extracting user message and adding it to state."""
        user_input = AGUIWorker._extract_user_input(run_input)

        logger.debug("Extracted latest user message", user_input=user_input[:100] if user_input else "(empty)")

        state_dict = dict(run_input.state) if run_input.state else {}
        state_dict["user_input"] = user_input

        return RunAgentInput(
            thread_id=run_input.thread_id,
            run_id=run_input.run_id,
            state=state_dict,
            messages=run_input.messages,
            tools=run_input.tools,
            context=run_input.context,
            forwarded_props=run_input.forwarded_props,
        )

    @staticmethod
    def _extract_user_input(run_input: RunAgentInput) -> str:
        """Extract the most recent user message from RunAgentInput messages."""
        for msg in reversed(run_input.messages):
            if msg.role == "user" and isinstance(msg.content, str):
                return msg.content
        return ""



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated
from uuid import UUID

from pydantic import ConfigDict, Field

from orchestrator.db.models import DESCRIPTION_LENGTH
from orchestrator.schemas.base import OrchestratorBaseModel


class ResourceTypeBaseSchema(OrchestratorBaseModel):
    resource_type: str
    description: str | None = None
    resource_type_id: UUID | None = None


class ResourceTypeSchema(ResourceTypeBaseSchema):
    resource_type_id: UUID
    model_config = ConfigDict(from_attributes=True)


class ResourceTypePatchSchema(OrchestratorBaseModel):
    description: Annotated[str, Field(max_length=DESCRIPTION_LENGTH)] | None = None



"""Add uuid search workaround.

Note: this workaround was added to existing migration da5c9f4cce1c in orchestrator-core commit 3e93263.
Because of that, it was never deployed to existing environments where the original migration was already executed.

This migration (460ec6748e37) will ensure the workaround is deployed onto existing environments.
The old migration (da5c9f4cce1c) is restored to its original state before commit 3e93263.

Revision ID: 460ec6748e37
Revises: 048219045729
Create Date: 2024-09-27 18:01:14.054599

"""

from pathlib import Path

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "460ec6748e37"
down_revision = "048219045729"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()

    revision_file_path = Path(__file__)
    with open(revision_file_path.with_suffix(".sql")) as f:
        conn.execute(sa.text(f.read()))


def downgrade() -> None:
    pass



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Callable

from pydantic_ai.toolsets import AbstractToolset

from orchestrator.search.agent.memory import MemoryScope
from orchestrator.search.agent.prompts import (
    get_aggregation_execution_prompt,
    get_result_actions_prompt,
    get_search_execution_prompt,
    get_text_response_prompt,
)
from orchestrator.search.agent.state import SearchState, TaskAction
from orchestrator.search.agent.tools import (
    aggregation_execution_toolset,
    aggregation_toolset,
    filter_building_toolset,
    result_actions_toolset,
    search_execution_toolset,
)


@dataclass(frozen=True)
class Skill:
    """Declarative capability definition (A2A sense, not Anthropic Agent Skills).

    In the A2A protocol, one agent advertises many skills — discrete
    capabilities with metadata (name, description, tags). Our Skill
    extends that with runtime wiring: toolsets, prompt function, and
    memory scope. The agent creates pydantic-ai Agent instances from
    skills at execution time.

    Note: Anthropic's "Agent Skills" are a different concept; filesystem-based
    instruction packages (SKILL.md) for teaching Claude workflows.
    """

    action: TaskAction
    name: str
    description: str
    tags: list[str]
    toolsets: list[AbstractToolset[Any]]
    get_prompt: Callable[[SearchState], str]
    memory_scope: MemoryScope


SKILLS: dict[TaskAction, Skill] = {
    TaskAction.SEARCH: Skill(
        action=TaskAction.SEARCH,
        name="Search",
        description="Find subscriptions, products, workflows, processes",
        tags=["search", "query"],
        toolsets=[filter_building_toolset, search_execution_toolset],
        get_prompt=get_search_execution_prompt,
        memory_scope=MemoryScope.LIGHTWEIGHT,
    ),
    TaskAction.AGGREGATION: Skill(
        action=TaskAction.AGGREGATION,
        name="Aggregate",
        description="Count, sum, avg with grouping",
        tags=["aggregate", "analytics"],
        toolsets=[filter_building_toolset, aggregation_toolset, aggregation_execution_toolset],
        get_prompt=get_aggregation_execution_prompt,
        memory_scope=MemoryScope.LIGHTWEIGHT,
    ),
    TaskAction.RESULT_ACTIONS: Skill(
        action=TaskAction.RESULT_ACTIONS,
        name="Result Actions",
        description="Export results or fetch entity details",
        tags=["export", "details"],
        toolsets=[result_actions_toolset],
        get_prompt=get_result_actions_prompt,
        memory_scope=MemoryScope.MINIMAL,
    ),
    TaskAction.TEXT_RESPONSE: Skill(
        action=TaskAction.TEXT_RESPONSE,
        name="Text Response",
        description="Answer general questions about the system",
        tags=["text", "help"],
        toolsets=[],
        get_prompt=get_text_response_prompt,
        memory_scope=MemoryScope.LIGHTWEIGHT,
    ),
}



# Copyright 2019-2024 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import cache
from typing import Any

import structlog

from orchestrator.db import ProductTable
from orchestrator.forms import SubmitFormPage
from orchestrator.forms.validators import Choice
from orchestrator.services.subscriptions import (
    get_subscriptions_on_product_table_in_sync,
)
from orchestrator.services.workflows import (
    get_validation_product_workflows_for_subscription,
    start_validation_workflow_for_workflows,
)
from orchestrator.settings import get_authorizers
from orchestrator.targets import Target
from orchestrator.workflow import StepList, done, init, step, workflow
from pydantic_forms.types import FormGenerator, State

authorizers = get_authorizers()
logger = structlog.get_logger(__name__)


def create_select_product_type_form() -> type[SubmitFormPage]:
    """Get and create the choices form for the product type."""

    @cache
    def get_product_type_choices() -> dict[Any, Any]:
        return {product.product_type: product.product_type for product in ProductTable.query.all()}

    ProductTypeChoices = Choice.__call__("Product Type", get_product_type_choices())

    class SelectProductTypeForm(SubmitFormPage):
        product_type: ProductTypeChoices  # type: ignore

    return SelectProductTypeForm


def initial_input_form_generator() -> FormGenerator:
    """Generate the form."""

    init_input = yield create_select_product_type_form()
    user_input_data = init_input.model_dump()

    return user_input_data


@step("Validate Product Type")
def validate_product_type(product_type: str) -> State:
    result = []
    subscriptions = get_subscriptions_on_product_table_in_sync()

    for subscription in subscriptions:
        system_product_workflows = get_validation_product_workflows_for_subscription(
            subscription=subscription,
        )

        if not system_product_workflows:
            logger.warning(
                "SubscriptionTable has no validation workflow",
                subscription=subscription,
                product=subscription.product.name,
            )
            continue

        validation_result = start_validation_workflow_for_workflows(
            subscription=subscription,
            workflows=system_product_workflows,
            product_type_filter=product_type,
        )
        if len(validation_result) > 0:
            result.append({"total_workflows_validated": len(validation_result), "workflows": validation_result})

    return {"result": result}


@workflow(
    "Validate all subscriptions of Product Type",
    target=Target.SYSTEM,
    initial_input_form=initial_input_form_generator,
    authorize_callback=authorizers.authorize_callback,
    retry_auth_callback=authorizers.retry_auth_callback,
)
def task_validate_product_type() -> StepList:
    return init >> validate_product_type >> done



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Query execution result models and formatting functions."""

from collections.abc import Sequence
from typing import Literal

from pydantic import BaseModel, ConfigDict, Field
from sqlalchemy.engine.row import RowMapping

from orchestrator.search.core.types import EntityType, FilterOp, SearchMetadata
from orchestrator.search.filters import FilterTree

from .queries import AggregateQuery, CountQuery, ExportQuery, SelectQuery


class VisualizationType(BaseModel):
    """Visualization type for aggregation results.

    Choose based on the query context:
    - 'pie': For categorical distributions (e.g., subscriptions by status, products by type)
    - 'line': For time-series data (e.g., subscriptions created per month, trends over time)
    - 'table': For detailed data or multiple grouping dimensions (default)
    """

    type: Literal["pie", "line", "table"] = Field(
        default="table",
        description="Visualization render type: 'pie' for categorical distributions, 'line' for time-series, 'table' for detailed data",
    )


class MatchingField(BaseModel):
    """Contains the field that contributed most to the (fuzzy) search result."""

    text: str
    path: str
    highlight_indices: list[tuple[int, int]] | None = None


class SearchResult(BaseModel):
    """Represents a single search result item."""

    entity_id: str
    entity_type: EntityType
    entity_title: str
    score: float
    perfect_match: int = 0
    matching_field: MatchingField | None = None


class SearchResponse(BaseModel):
    """Response containing search results and metadata."""

    results: list[SearchResult]
    metadata: SearchMetadata
    query_embedding: list[float] | None = None
    has_more: bool = False
    total_items: int | None = None
    start_cursor: int | None = None
    end_cursor: int | None = None


class ResultRow(BaseModel):
    """A single result row with labeled string columns and numeric columns."""

    group_values: dict[str, str] = Field(default_factory=dict)
    aggregations: dict[str, float | int] = Field(default_factory=dict)

    model_config = ConfigDict(extra="forbid")


class QueryResultsResponse(BaseModel):
    """Tabular query results used for both search and aggregation rendering."""

    results: list[ResultRow]
    total_results: int
    metadata: SearchMetadata
    visualization_type: VisualizationType = Field(default_factory=VisualizationType)

    model_config = ConfigDict(extra="forbid")


class ExportData(BaseModel):
    """Export metadata for download."""

    action: str = "export"
    query_id: str
    download_url: str
    message: str


def format_aggregation_response(
    result_rows: Sequence[RowMapping],
    group_column_names: list[str],
    query: CountQuery | AggregateQuery,
) -> QueryResultsResponse:
    """Format raw aggregation query results into QueryResultsResponse.

    Args:
        result_rows: Raw database result rows
        group_column_names: List of column names that are grouping columns
        query: Query plan for metadata

    Returns:
        QueryResultsResponse with formatted results and metadata
    """
    results = []
    for row in result_rows:
        group_values = {}
        aggregations = {}

        for key, value in row.items():
            if key in group_column_names:
                group_values[key] = str(value) if value is not None else ""
            else:
                aggregations[key] = value if value is not None else 0

        results.append(ResultRow(group_values=group_values, aggregations=aggregations))

    metadata = SearchMetadata(
        search_type="aggregation",
        description=f"Aggregation query with {len(query.group_by or [])} grouping dimension(s)",
    )

    return QueryResultsResponse(
        results=results,
        total_results=len(results),
        metadata=metadata,
    )


def truncate_text_with_highlights(
    text: str, highlight_indices: list[tuple[int, int]] | None = None, max_length: int = 500, context_chars: int = 100
) -> tuple[str, list[tuple[int, int]] | None]:
    """Truncate text to max_length while preserving context around the first highlight.

    Args:
        text: The text to truncate
        highlight_indices: List of (start, end) tuples indicating highlight positions, or None
        max_length: Maximum length of the returned text
        context_chars: Number of characters to show before and after the first highlight

    Returns:
        Tuple of (truncated_text, adjusted_highlight_indices)
    """
    # If text is short enough, return as-is
    if len(text) <= max_length:
        return text, highlight_indices

    # If no highlights, truncate from beginning
    if not highlight_indices:
        truncated_text = text[:max_length]
        suffix = "..." if len(text) > max_length else ""
        return truncated_text + suffix, None

    # Use first highlight to determine what to show
    first_highlight_start = highlight_indices[0][0]

    # Calculate start position: try to center around first highlight
    start = max(0, first_highlight_start - context_chars)
    end = min(len(text), start + max_length)

    # Adjust start if we hit the end boundary
    if end == len(text) and (end - start) < max_length:
        start = max(0, end - max_length)

    truncated_text = text[start:end]

    # Add ellipsis to indicate truncation
    truncated_from_start = start > 0
    truncated_from_end = end < len(text)

    if truncated_from_start:
        truncated_text = "..." + truncated_text
    if truncated_from_end:
        truncated_text = truncated_text + "..."

    # Adjust highlight indices to be relative to truncated text
    offset = start - (3 if truncated_from_start else 0)  # Account for leading "..."
    adjusted_indices = []
    for hl_start, hl_end in highlight_indices:
        # Only include highlights that are within the truncated range
        if hl_start >= start and hl_end <= end:
            adjusted_indices.append((hl_start - offset, hl_end - offset))

    return truncated_text, adjusted_indices if adjusted_indices else None


def generate_highlight_indices(text: str, term: str) -> list[tuple[int, int]]:
    """Finds all occurrences of individual words from the term, including both word boundary and substring matches."""
    import re

    if not text or not term:
        return []

    all_matches = []
    words = [w.strip() for w in term.split() if w.strip()]

    for word in words:
        # First find word boundary matches
        word_boundary_pattern = rf"\b{re.escape(word)}\b"
        word_matches = list(re.finditer(word_boundary_pattern, text, re.IGNORECASE))
        all_matches.extend([(m.start(), m.end()) for m in word_matches])

        # Then find all substring matches
        substring_pattern = re.escape(word)
        substring_matches = list(re.finditer(substring_pattern, text, re.IGNORECASE))
        all_matches.extend([(m.start(), m.end()) for m in substring_matches])

    return sorted(set(all_matches))


def format_search_response(
    db_rows: Sequence[RowMapping],
    query: "SelectQuery | ExportQuery",
    metadata: SearchMetadata,
    query_embedding: list[float] | None,
    total_items: int | None,
    start_cursor: int | None,
    end_cursor: int | None,
) -> SearchResponse:
    """Format database query results into a `SearchResponse`.

    Converts raw SQLAlchemy `RowMapping` objects into `SearchResult` instances,
    including highlight metadata if present in the database results.

    Args:
        db_rows: The rows returned from the executed SQLAlchemy query.
        query: SelectQuery or ExportQuery with search criteria.
        metadata: Metadata about the search execution.
        query_embedding: query embedding for agent to save to DB.
        total_items: total items of the query (only with structured search).
        start_cursor: start cursor of the db_rows (only with structured search).
        end_cursor: end cursor of the db_rows (only with structured search).

    Returns:
        SearchResponse: A list of `SearchResult` objects containing entity IDs, scores,
        and optional highlight information.
    """
    from orchestrator.search.retrieval.retrievers import Retriever

    if not db_rows:
        return SearchResponse(results=[], metadata=metadata)

    user_query = query.query_text

    results = []
    for row in db_rows:
        matching_field = None

        if (
            user_query
            and (text := row.get(Retriever.HIGHLIGHT_TEXT_LABEL)) is not None
            and (path := row.get(Retriever.HIGHLIGHT_PATH_LABEL)) is not None
        ):
            if not isinstance(text, str):
                text = str(text)
            if not isinstance(path, str):
                path = str(path)

            highlight_indices = generate_highlight_indices(text, user_query)
            truncated_text, adjusted_indices = truncate_text_with_highlights(text, highlight_indices)
            matching_field = MatchingField(text=truncated_text, path=path, highlight_indices=adjusted_indices)

        elif not user_query and query.filters and metadata.search_type == "structured":
            # Structured search (filter-only)
            matching_field = _extract_matching_field_from_filters(query.filters)

        entity_title = row.get("entity_title", "")
        if not isinstance(entity_title, str):
            entity_title = str(entity_title) if entity_title is not None else ""

        results.append(
            SearchResult(
                entity_id=str(row.entity_id),
                entity_type=query.entity_type,
                entity_title=entity_title,
                score=row.score,
                perfect_match=row.get("perfect_match", 0),
                matching_field=matching_field,
            )
        )
    return SearchResponse(
        results=results,
        metadata=metadata,
        total_items=total_items,
        start_cursor=start_cursor,
        end_cursor=end_cursor,
        query_embedding=query_embedding,
    )


def _extract_matching_field_from_filters(filters: "FilterTree") -> MatchingField | None:
    """Extract the first path filter to use as matching field for structured searches."""
    from orchestrator.search.filters import LtreeFilter

    leaves = filters.get_all_leaves()
    if len(leaves) != 1:
        return None

    pf = leaves[0]

    if isinstance(pf.condition, LtreeFilter):
        op = pf.condition.op
        # Prefer the original component/pattern (validator may set path="*" and move the value)
        display = str(getattr(pf.condition, "value", "") or pf.path)

        # There can be no match for abscence.
        if op == FilterOp.NOT_HAS_COMPONENT:
            return None

        return MatchingField(text=display, path=display, highlight_indices=[(0, len(display))])

    # Everything thats not Ltree
    val = getattr(pf.condition, "value", "")
    text = "" if val is None else str(val)
    return MatchingField(text=text, path=pf.path, highlight_indices=[(0, len(text))])



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from uuid import UUID

import structlog
from sqlalchemy import select

from orchestrator.db import ProcessTable, db
from orchestrator.services import processes
from orchestrator.settings import get_authorizers
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus, StepList, done, init, step, workflow
from pydantic_forms.types import State, UUIDstr

authorizers = get_authorizers()
logger = structlog.get_logger(__name__)


def get_process_ids_by_process_statuses(process_statuses: list[ProcessStatus], exclude_ids: list[UUID]) -> list:
    return list(
        db.session.scalars(
            select(ProcessTable.process_id).filter(
                ProcessTable.last_status.in_(process_statuses), ProcessTable.process_id.not_in(exclude_ids)
            )
        )
    )


@step("Find waiting workflows")
def find_waiting_workflows(process_id: UUID) -> State:
    created_process_ids = get_process_ids_by_process_statuses([ProcessStatus.CREATED], exclude_ids=[process_id])
    resumed_process_ids = get_process_ids_by_process_statuses([ProcessStatus.RESUMED], exclude_ids=[process_id])
    waiting_process_ids = get_process_ids_by_process_statuses([ProcessStatus.WAITING], exclude_ids=[process_id])

    return {
        "number_of_waiting_processes": len(waiting_process_ids),
        "waiting_process_ids": waiting_process_ids,
        "created_processes_stuck": len(created_process_ids),
        "created_state_process_ids": created_process_ids,
        "resumed_processes_stuck": len(resumed_process_ids),
        "resumed_state_process_ids": resumed_process_ids,
    }


@step("Resume found workflows")
def resume_found_workflows(
    waiting_process_ids: list[UUIDstr],
    resumed_state_process_ids: list[UUIDstr],
) -> State:
    resume_processes = waiting_process_ids + resumed_state_process_ids

    resumed_process_ids: list = []
    for process_id in resume_processes:
        try:
            process = db.session.get(ProcessTable, process_id)
            if not process:
                continue

            # Workaround the commit disable function
            db.session.info["disabled"] = False

            processes.resume_process(process)
            resumed_process_ids.append(process_id)
        except Exception as exc:
            logger.warning("Could not resume process", process_id=process_id, error=str(exc))
        finally:
            # Make sure to turn it on again
            db.session.info["disabled"] = True

    return {
        "number_of_resumed_process_ids": len(resumed_process_ids),
        "resumed_process_ids": resumed_process_ids,
    }


@step("Restart found CREATED workflows")
def restart_created_workflows(created_state_process_ids: list[UUIDstr]) -> State:
    started_process_ids = []
    for process_id in created_state_process_ids:
        try:
            process = db.session.get(ProcessTable, process_id)
            if not process:
                continue

            # Workaround the commit disable function
            db.session.info["disabled"] = False

            processes.restart_process(process)
            started_process_ids.append(process_id)
        except Exception as exc:
            logger.warning("Could not resume process", process_id=process_id, error=str(exc))
        finally:
            # Make sure to turn it on again
            db.session.info["disabled"] = True

    return {
        "number_of_started_process_ids": len(started_process_ids),
        "started_process_ids": started_process_ids,
    }


@workflow(
    "Resume all workflows that are stuck on tasks with the status 'waiting', 'created' or 'resumed'",
    target=Target.SYSTEM,
    authorize_callback=authorizers.authorize_callback,
    retry_auth_callback=authorizers.retry_auth_callback,
)
def task_resume_workflows() -> StepList:
    return init >> find_waiting_workflows >> resume_found_workflows >> restart_created_workflows >> done



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
from typing import Annotated, Any

import structlog
from fastapi import Depends
from fastapi.routing import APIRouter

from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.api.error_handling import raise_status
from orchestrator.security import authenticate
from pydantic_forms.core.asynchronous import start_form
from pydantic_forms.exceptions import FormException

logger = structlog.get_logger(__name__)

router: APIRouter = APIRouter()


@router.post("/{form_key}", status_code=HTTPStatus.OK)
async def new_form(
    form_key: str,
    json_data: list[dict[str, Any]],
    user_model: Annotated[OIDCUserModel | None, Depends(authenticate)],
) -> dict[str, Any]:
    username = user_model.user_name if user_model else ""
    try:
        return await start_form(form_key, user_inputs=json_data, user=username, user_model=user_model)
    except FormException as exc:
        if "does not exist" in str(exc):
            return raise_status(HTTPStatus.NOT_FOUND, str(exc))
        raise exc



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Annotated

from pydantic import Field

from orchestrator.db import db
from orchestrator.db.models import NOTE_LENGTH
from orchestrator.forms import SubmitFormPage
from orchestrator.services import subscriptions
from orchestrator.settings import get_authorizers
from orchestrator.targets import Target
from orchestrator.utils.json import to_serializable
from orchestrator.workflow import StepList, done, init, step, workflow
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic_forms.types import FormGenerator, State, UUIDstr
from pydantic_forms.validators import LongText

authorizers = get_authorizers()


def initial_input_form(subscription_id: UUIDstr) -> FormGenerator:
    subscription = subscriptions.get_subscription(subscription_id)
    subscription_backup = {subscription_id: to_serializable(subscription)}
    old_note = subscription.note

    class ModifyNoteForm(SubmitFormPage):
        note: Annotated[LongText, Field(max_length=NOTE_LENGTH)] = old_note

    user_input = yield ModifyNoteForm

    return {
        "old_note": old_note,
        "note": user_input.note,
        "__old_subscriptions__": subscription_backup,
    }


@step("Store note")
def store_subscription_note(subscription_id: UUIDstr, note: str) -> State:
    subscription = subscriptions.get_subscription(subscription_id)

    subscription.note = note
    db.session.add(subscription)

    return {
        "subscription": to_serializable(subscription),
    }


@workflow(
    "Modify Note",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form),
    target=Target.MODIFY,
    authorize_callback=authorizers.authorize_callback,
    retry_auth_callback=authorizers.retry_auth_callback,
)
def modify_note() -> StepList:
    return init >> store_process_subscription() >> store_subscription_note >> done



from typing import Any, TypeVar, cast

T = TypeVar("T")


def get_original_model(model: Any, klass: T) -> T:
    """Get original type in a typesafe way."""
    original_model = getattr(model, "_original_model", None)

    if original_model:
        return cast(T, original_model)
    raise ValueError(f"Cant get original model for type {klass}")



"""Add workflow_id to processes table.

Revision ID: 048219045729
Revises: da5c9f4cce1c
Create Date: 2023-12-06 15:33:46.997517

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.exc import IntegrityError

from orchestrator import db

# revision identifiers, used by Alembic.
revision = "048219045729"
down_revision = "da5c9f4cce1c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Add new processes.workflow_id column
    op.add_column("processes", column=sa.Column("workflow_id", sqlalchemy_utils.types.uuid.UUIDType()))
    # Fill workflow_id column based on the workflow name
    op.execute(
        "UPDATE processes SET workflow_id = wf.workflow_id FROM workflows wf WHERE processes.workflow = wf.name;"
    )

    try:
        op.alter_column("processes", "workflow_id", nullable=False)
    except IntegrityError:
        raise Exception(
            """
        Migration failed due to processes.workflow rows that have a workflow name that no longer exists.

        After this update, each process must be linked to a workflow.

        Follow these steps:
        1) Make a backup of the processes that are not linked to a workflow (i.e. processes.workflow does match a workflows.name).
        2) Manually delete the rows in processes that are not linked to a workflow.
        3) Re-run this migration.
        """
        )

    op.execute(
        """ALTER TABLE processes
        ADD CONSTRAINT processes_workflow_id_fkey FOREIGN KEY (workflow_id) REFERENCES workflows (workflow_id);"""
    )
    op.drop_column("processes", "workflow")

    # Add deleted_at column to workflows table
    op.add_column("workflows", column=sa.Column("deleted_at", db.UtcTimestamp(timezone=True)))


def downgrade() -> None:
    op.add_column(
        "processes",
        column=sa.Column("workflow", sa.String, nullable=False, server_default=""),
    )
    op.execute(
        "UPDATE processes SET workflow = wf.name FROM workflows wf WHERE processes.workflow_id = wf.workflow_id;"
    )
    op.drop_constraint("processes_workflow_id_fkey", table_name="processes")
    op.drop_column("processes", column_name="workflow_id")
    op.drop_column("workflows", column_name="deleted_at")



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import uuid
from contextlib import AsyncExitStack, asynccontextmanager
from types import TracebackType
from typing import AsyncIterator, Sequence, cast

import structlog
from fasta2a.applications import FastA2A
from fasta2a.broker import InMemoryBroker
from fasta2a.schema import Message, Part, Skill, TaskSendParams
from fasta2a.schema import TextPart as A2ATextPart
from fasta2a.storage import InMemoryStorage
from pydantic_ai._a2a import AgentWorker
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.messages import (
    ModelRequest,
    ModelResponse,
    UserPromptPart,
)
from pydantic_ai.messages import (
    TextPart as AiTextPart,
)

from orchestrator.search.agent.adapters.stream import collect_stream_output
from orchestrator.search.agent.agent import AgentAdapter
from orchestrator.search.agent.skills import SKILLS
from orchestrator.search.agent.state import SearchState, TaskAction

logger = structlog.get_logger(__name__)

A2A_SKILLS = [
    Skill(
        id=action.value,
        name=skill.name,
        description=skill.description,
        tags=skill.tags,
        input_modes=["application/json"],
        output_modes=["application/json"],
    )
    for action, skill in SKILLS.items()
]


class A2AWorker(AgentWorker):
    """AgentWorker subclass that consumes the agent's event stream directly.

    Bypasses ``AgentWorker.run_task()`` (which calls ``agent.run()``) and
    instead drives ``agent.run_stream_events()`` — the same pipeline used by
    the AG-UI adapter.  This keeps ``AgentAdapter`` protocol-agnostic while
    giving the A2A adapter full control over stream consumption and result
    assembly.

    The A2A protocol advertises skills on the Agent Card but has no first-class
    field for targeting a skill on ``message/send``. The convention is to pass
    ``{"skill_id": "<action>"}`` in the message metadata.  This worker extracts
    that hint and passes it as ``target_action`` to skip the planner.
    """

    async def run_task(self, params: TaskSendParams) -> None:
        agent = cast(AgentAdapter, self.agent)

        task = await self.storage.load_task(params["id"])
        if task is None:
            raise ValueError(f'Task {params["id"]} not found')

        if task["status"]["state"] != "submitted":
            raise ValueError(f'Task {params["id"]} has already been processed (state: {task["status"]["state"]})')

        await self.storage.update_task(task["id"], state="working")

        metadata = params["message"].get("metadata", {}) or {}
        skill_id = metadata.get("skill_id") or metadata.get("skillId")
        target_action: TaskAction | None = None

        if skill_id:
            try:
                target_action = TaskAction(skill_id)
                logger.debug("A2A: Routing to skill directly", target_action=target_action)
            except ValueError:
                logger.warning("A2A: Unknown skillId, falling back to planner", skill_id=skill_id)

        user_input = self._extract_user_input(task.get("history", []))

        deps = StateDeps(SearchState(user_input=user_input))

        from orchestrator.db import db
        from orchestrator.db.models import AgentRunTable

        deps.state.run_id = uuid.uuid4()
        agent_run = AgentRunTable(run_id=deps.state.run_id, thread_id=str(uuid.uuid4()), agent_type="a2a")
        db.session.add(agent_run)
        db.session.commit()

        logger.debug("A2AWorker: Starting execution", task_id=task["id"])

        try:
            event_stream = agent.run_stream_events(deps=deps, target_action=target_action)
            final_output = await collect_stream_output(event_stream)

            a2a_messages: list[Message] = [
                Message(
                    role="agent",
                    parts=[A2ATextPart(kind="text", text=final_output)],
                    kind="message",
                    message_id=str(uuid.uuid4()),
                )
            ]

            # Update context with synthetic message history so multi-turn works
            context_messages = await self.storage.load_context(task["context_id"]) or []
            context_messages.extend(
                [
                    ModelRequest(parts=[UserPromptPart(content=user_input)]),
                    ModelResponse(parts=[AiTextPart(content=final_output)]),
                ]
            )
            await self.storage.update_context(task["context_id"], context_messages)

            artifacts = self.build_artifacts(final_output)
        except Exception:
            await self.storage.update_task(task["id"], state="failed")
            raise
        else:
            await self.storage.update_task(
                task["id"], state="completed", new_artifacts=artifacts, new_messages=a2a_messages
            )

    @staticmethod
    def _extract_user_input(history: list[Message] | Sequence[Message]) -> str:
        """Extract user input text from A2A task history messages."""

        def _is_text_part(part: Part) -> bool:
            return part.get("kind") == "text"

        for msg in reversed(history):
            if msg.get("role") == "user":
                for part in msg.get("parts", []):
                    if _is_text_part(part):
                        return cast(A2ATextPart, part)["text"]
        return ""


class A2AApp:
    """A2A adapter app: FastA2A server, worker, and lifecycle.

    Builds the FastA2A app, broker, storage, and worker so that the worker
    lifecycle can be tied to the host application's lifespan when mounted
    as a sub-app.
    """

    def __init__(self, agent: AgentAdapter, url: str = "") -> None:
        self.agent = agent

        storage = InMemoryStorage()
        broker = InMemoryBroker()
        self.worker = A2AWorker(agent=agent, broker=broker, storage=storage)  # type: ignore[arg-type]

        @asynccontextmanager
        async def _noop_lifespan(_app: FastA2A) -> AsyncIterator[None]:
            yield

        self.app = FastA2A(
            storage=storage,
            broker=broker,
            name="WFO Search Agent",
            url=url,
            description="Search, filter and aggregate orchestration data",
            skills=A2A_SKILLS,
            lifespan=_noop_lifespan,
        )

    async def __aenter__(self) -> A2AApp:
        self._stack = AsyncExitStack()
        await self._stack.__aenter__()
        await self._stack.enter_async_context(self.app.task_manager)
        await self._stack.enter_async_context(self.agent)
        await self._stack.enter_async_context(self.worker.run())
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._stack.__aexit__(exc_type, exc_val, exc_tb)



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import strawberry

from pydantic_forms.types import strEnum


@strawberry.enum
class Target(strEnum):
    CREATE = "CREATE"
    MODIFY = "MODIFY"
    TERMINATE = "TERMINATE"
    SYSTEM = "SYSTEM"
    VALIDATE = "VALIDATE"
    RECONCILE = "RECONCILE"



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nwastdlib.vlans import VlanRanges
from orchestrator.forms.validators.customer_contact_list import customer_contact_list
from orchestrator.forms.validators.customer_id import CustomerId
from orchestrator.forms.validators.display_subscription import DisplaySubscription
from orchestrator.forms.validators.product_id import ProductId, ProductIdError, product_id
from pydantic_forms.types import strEnum
from pydantic_forms.validators import (
    Accept,
    Choice,
    ContactPerson,
    ContactPersonName,
    Divider,
    Label,
    ListOfOne,
    ListOfTwo,
    LongText,
    MigrationSummary,
    OrganisationId,
    Timestamp,
    choice_list,
    contact_person_list,
    migration_summary,
    timestamp,
    unique_conlist,
)
from pydantic_forms.validators.helpers import remove_empty_items

__all__ = [
    "Accept",
    "Choice",
    "choice_list",
    "ContactPerson",
    "ContactPersonName",
    "contact_person_list",
    "CustomerId",
    "DisplaySubscription",
    "Divider",
    "Label",
    "ListOfOne",
    "ListOfTwo",
    "LongText",
    "ProductIdError",
    "ProductId",
    "MigrationSummary",
    "OrganisationId",
    "Timestamp",
    "migration_summary",
    "product_id",
    "remove_empty_items",
    "strEnum",
    "timestamp",
    "unique_conlist",
    "VlanRanges",
    "customer_contact_list",
]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cache
from typing import Annotated

from ag_ui.core import RunAgentInput
from fastapi import APIRouter, Depends, HTTPException, Request
from pydantic import ValidationError
from pydantic_ai.ag_ui import SSE_CONTENT_TYPE
from starlette.responses import Response, StreamingResponse
from structlog import get_logger

from orchestrator.db import db
from orchestrator.search.agent.adapters import AGUIWorker
from orchestrator.search.agent.agent import AgentAdapter

router = APIRouter()
logger = get_logger(__name__)


@cache
def get_agent(request: Request) -> AgentAdapter:
    """Dependency to provide the agent instance.

    The agent is built once and cached for the lifetime of the application.
    """
    from orchestrator.llm_settings import llm_settings

    model = request.app.agent_model
    debug = llm_settings.AGENT_DEBUG

    return AgentAdapter(model, debug=debug)


@router.post("/")
async def agent_conversation(
    request: Request,
    agent: Annotated[AgentAdapter, Depends(get_agent)],
) -> Response:
    """Agent conversation endpoint using pydantic-ai AG-UI protocol."""
    try:
        body = await request.json()
        run_input = RunAgentInput(**body)
    except ValidationError as e:
        logger.error("Invalid request body", error=str(e))
        raise HTTPException(status_code=400, detail=f"Invalid request body: {e}")

    stream = await AGUIWorker.run_request(agent=agent, run_input=run_input, db_session=db.session)
    return StreamingResponse(stream, media_type=SSE_CONTENT_TYPE)



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import UUID

from sqlalchemy import UUID as SA_UUID
from sqlalchemy import cast as sa_cast
from sqlalchemy import select
from sqlalchemy.orm import raiseload

from orchestrator.db import SubscriptionInstanceRelationTable, SubscriptionInstanceTable, db


def _eagerload_subscription_instances(
    subscription_id: UUID | str, instance_attributes: list[str]
) -> list[SubscriptionInstanceTable]:
    """Given a subscription id, recursively query all depends_on subscription instances with the instance_attributes eagerloaded.

    Note: accessing instance attributes on the result that were not explicitly loaded will
    trigger a sqlalchemy error.
    """
    from orchestrator.db.loaders import get_query_loaders_for_model_paths

    # CTE to recursively get all subscription instance ids the subscription depends on
    instance_ids_cte = (
        select(
            sa_cast(None, SA_UUID(as_uuid=True)).label("in_use_by_id"),
            SubscriptionInstanceTable.subscription_instance_id.label("depends_on_id"),
        )
        .where(SubscriptionInstanceTable.subscription_id == subscription_id)
        .cte(name="recursive_instance_ids", recursive=True)
    )

    cte_alias = instance_ids_cte.alias()
    rel_alias = select(SubscriptionInstanceRelationTable).alias()

    instance_ids = instance_ids_cte.union(
        select(rel_alias.c.in_use_by_id, rel_alias.c.depends_on_id).where(
            rel_alias.c.in_use_by_id == cte_alias.c.depends_on_id
        )
    )

    select_all_instance_ids = select(instance_ids.c.depends_on_id).subquery()

    # Eagerload specified instance attributes
    query_loaders = get_query_loaders_for_model_paths(SubscriptionInstanceTable, instance_attributes)
    # Prevent unwanted lazyloading of all other attributes
    query_loaders += [raiseload("*")]  # type: ignore[list-item]  # todo fix this type
    stmt = (
        select(SubscriptionInstanceTable)
        .where(SubscriptionInstanceTable.subscription_instance_id.in_(select(select_all_instance_ids)))
        .options(*query_loaders)
    )

    return db.session.scalars(stmt).all()  # type: ignore[return-value]  # todo fix this type


def eagerload_all_subscription_instances(subscription_id: UUID | str) -> list[SubscriptionInstanceTable]:
    """Recursively find the subscription's depends_on instances and resolve relations for SubscriptionModel.from_subscription_id()."""
    instance_attributes = [
        "subscription.product",
        "product_block",
        "values.resource_type",
        "depends_on",
        "in_use_by",
    ]
    return _eagerload_subscription_instances(subscription_id, instance_attributes)


def eagerload_all_subscription_instances_only_inuseby(subscription_id: UUID | str) -> list[SubscriptionInstanceTable]:
    """Recursively find the subscription's depends_on instances and resolve their in_use_by relations."""
    instance_attributes = [
        "in_use_by",
    ]
    return _eagerload_subscription_instances(subscription_id, instance_attributes)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, cast

from structlog import get_logger

from orchestrator.db.database import BaseModel as DbBaseModel
from orchestrator.db.database import Database, transactional
from orchestrator.db.models import (  # noqa: F401
    AgentRunTable,
    EngineSettingsTable,
    FixedInputTable,
    InputStateTable,
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SearchQueryTable,
    SubscriptionCustomerDescriptionTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionMetadataTable,
    SubscriptionTable,
    UtcTimestamp,
    UtcTimestampError,
    WorkflowTable,
)
from orchestrator.settings import AppSettings

logger = get_logger(__name__)


class WrappedDatabase:
    def __init__(self, wrappee: Database | None = None) -> None:
        self.wrapped_database = wrappee

    def update(self, wrappee: Database) -> None:
        self.wrapped_database = wrappee
        logger.info("Database object configured, all methods referencing `db` should work.")

    def __getattr__(self, attr: str) -> Any:
        if not isinstance(self.wrapped_database, Database):
            if "_" in attr:
                logger.warning("No database configured, but attempting to access class methods")
                return None
            raise RuntimeWarning(
                "No database configured at this time. Please pass database configuration to OrchestratorCore base_settings"
            )

        return getattr(self.wrapped_database, attr)


# You need to pass a modified AppSettings class to the OrchestratorCore class to init the database correctly
wrapped_db = WrappedDatabase()
db = cast(Database, wrapped_db)


# The Global Database is set after calling this function
def init_database(settings: AppSettings) -> Database:
    wrapped_db.update(Database(str(settings.DATABASE_URI.get_secret_value())))
    return db


__all__ = [
    "transactional",
    "SearchQueryTable",
    "AgentRunTable",
    "SubscriptionTable",
    "ProcessSubscriptionTable",
    "ProcessTable",
    "ProcessStepTable",
    "ProductTable",
    "ProductBlockTable",
    "SubscriptionInstanceRelationTable",
    "SubscriptionInstanceTable",
    "SubscriptionInstanceValueTable",
    "SubscriptionMetadataTable",
    "ResourceTypeTable",
    "FixedInputTable",
    "InputStateTable",
    "EngineSettingsTable",
    "WorkflowTable",
    "SubscriptionCustomerDescriptionTable",
    "UtcTimestamp",
    "UtcTimestampError",
    "db",
    "init_database",
]

ALL_DB_MODELS: list[type[DbBaseModel]] = [
    SearchQueryTable,
    AgentRunTable,
    FixedInputTable,
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionCustomerDescriptionTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionMetadataTable,
    SubscriptionTable,
    WorkflowTable,
]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated, Any, ClassVar, Literal, Self, Union

from pydantic import BaseModel, ConfigDict, Discriminator, Field, TypeAdapter, model_validator

from orchestrator.search.core.types import EntityType, QueryOperation
from orchestrator.search.filters import FilterTree

from .mixins import (
    AggregationMixin,
    GroupingMixin,
    SearchMixin,
    StructuredOrderByMixin,
)


class BaseQuery(BaseModel):
    """Base model for all query types.

    Contains shared constants, properties, and utilities.
    """

    MIN_LIMIT: ClassVar[int] = 1
    DEFAULT_LIMIT: ClassVar[int] = 10
    MAX_LIMIT: ClassVar[int] = 30
    DEFAULT_EXPORT_LIMIT: ClassVar[int] = 1000
    MAX_EXPORT_LIMIT: ClassVar[int] = 10000

    _query_operation: ClassVar[QueryOperation]

    entity_type: EntityType
    filters: FilterTree | None = Field(default=None, description="Structured filters to apply")

    model_config = ConfigDict(extra="forbid")

    @property
    def query_operation(self) -> QueryOperation:
        return self._query_operation

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        """Build query from a dictionary.

        Args:
            data: Dictionary with query parameters

        Returns:
            Query instance of the appropriate type
        """
        return cls.model_validate(data)


class SelectQuery(BaseQuery, SearchMixin, StructuredOrderByMixin):
    """Query for SELECT operations.

    Composes BaseQuery with SearchMixin for text search, with strict result limits.
    """

    query_type: Literal["select"] = "select"
    _query_operation: ClassVar[QueryOperation] = QueryOperation.SELECT

    limit: int = Field(
        default=BaseQuery.DEFAULT_LIMIT,
        ge=BaseQuery.MIN_LIMIT,
        le=BaseQuery.MAX_LIMIT,
        description="Maximum number of search results to return",
    )


class ExportQuery(BaseQuery, SearchMixin, StructuredOrderByMixin):
    """Query for EXPORT operations .

    Similar to SelectQuery but with higher limits for bulk exports.
    """

    query_type: Literal["export"] = "export"
    _query_operation: ClassVar[QueryOperation] = QueryOperation.SELECT

    limit: int = Field(
        default=BaseQuery.DEFAULT_EXPORT_LIMIT,
        ge=BaseQuery.MIN_LIMIT,
        le=BaseQuery.MAX_EXPORT_LIMIT,
        description="Maximum number of results to export",
    )


class CountQuery(BaseQuery, GroupingMixin):
    """Query for COUNT operations with optional grouping."""

    query_type: Literal["count"] = "count"
    _query_operation: ClassVar[QueryOperation] = QueryOperation.COUNT


class AggregateQuery(BaseQuery, GroupingMixin, AggregationMixin):
    """Query for AGGREGATE operations.

    Composes BaseQuery with GroupingMixin and AggregationMixin
    to provide both grouping and aggregation capabilities.
    """

    query_type: Literal["aggregate"] = "aggregate"
    _query_operation: ClassVar[QueryOperation] = QueryOperation.AGGREGATE

    @model_validator(mode="after")
    def validate_cumulative_aggregation_types(self) -> Self:
        """Validate that cumulative is only used with COUNT and SUM aggregations."""
        if self.cumulative:
            from orchestrator.search.aggregations import AggregationType

            for agg in self.aggregations:
                if agg.type in (AggregationType.AVG, AggregationType.MIN, AggregationType.MAX):
                    raise ValueError(
                        f"Cumulative aggregations are not supported for {agg.type.value.upper()} aggregations. "
                        f"Cumulative only works with COUNT and SUM."
                    )
        return self

    def get_pivot_fields(self) -> list[str]:
        """Get all fields needed for EAV pivot including aggregation fields."""
        # Get grouping fields from GroupingMixin
        fields = super().get_pivot_fields()

        # Add aggregation fields
        fields.extend(self.get_aggregation_pivot_fields())

        return list(dict.fromkeys(fields))


Query = Annotated[
    Union[SelectQuery, ExportQuery, CountQuery, AggregateQuery],
    Discriminator("query_type"),
]

QueryAdapter: TypeAdapter[Query] = TypeAdapter(Query)



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus

import structlog
from fastapi import Depends
from fastapi.routing import APIRouter
from more_itertools import first

from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.api.error_handling import raise_status
from orchestrator.schedules.service import add_scheduled_task_to_queue, get_linker_entries_by_schedule_ids
from orchestrator.schemas.schedules import APSchedulerJobCreate, APSchedulerJobDelete, APSchedulerJobUpdate
from orchestrator.security import authenticate
from orchestrator.services.workflows import get_workflow_by_workflow_id
from orchestrator.workflow import Workflow
from orchestrator.workflows import get_workflow

logger = structlog.get_logger(__name__)

router: APIRouter = APIRouter()


async def validate_schedule_authorization(task: Workflow | None, user_model: OIDCUserModel | None) -> None:
    if not task:
        raise_status(HTTPStatus.NOT_FOUND, "Task does not exist")
    if not await task.authorize_callback(user_model):
        raise_status(HTTPStatus.FORBIDDEN, f"User is not authorized to manage schedule with '{task.name}' task")


@router.post("/", status_code=HTTPStatus.CREATED)
async def create_scheduled_task(
    payload: APSchedulerJobCreate, user_model: OIDCUserModel | None = Depends(authenticate)
) -> dict[str, str]:
    """Create a scheduled task."""
    task_key = payload.workflow_name
    task = get_workflow(task_key)

    await validate_schedule_authorization(task, user_model)
    add_scheduled_task_to_queue(payload)
    return {"message": "Added to Create Queue", "status": "CREATED"}


@router.put("/", status_code=HTTPStatus.OK)
async def update_scheduled_task(
    payload: APSchedulerJobUpdate, user_model: OIDCUserModel | None = Depends(authenticate)
) -> dict[str, str]:
    """Update a scheduled task."""
    schedules = get_linker_entries_by_schedule_ids([str(payload.schedule_id)])
    if not (schedule := first(schedules, None)):
        raise_status(HTTPStatus.NOT_FOUND, "Schedule does not exist")
    if not (workflow_table := get_workflow_by_workflow_id(str(schedule.workflow_id))):
        raise_status(HTTPStatus.NOT_FOUND, "Task does not exist")

    task_key = workflow_table.name
    task = get_workflow(task_key)

    await validate_schedule_authorization(task, user_model)
    add_scheduled_task_to_queue(payload)
    return {"message": "Added to Update Queue", "status": "UPDATED"}


@router.delete("/", status_code=HTTPStatus.OK)
async def delete_scheduled_task(payload: APSchedulerJobDelete) -> dict[str, str]:
    """Delete a scheduled task."""
    add_scheduled_task_to_queue(payload)
    return {"message": "Added to Delete Queue", "status": "DELETED"}



from uuid import UUID

import strawberry
from strawberry.scalars import JSON

from oauth2_lib.strawberry import authenticated_mutation_field
from orchestrator.api.api_v1.endpoints.processes import resolve_user_name
from orchestrator.graphql.types import MutationError, OrchestratorInfo
from orchestrator.services.process_broadcast_thread import graphql_broadcast_process_data
from orchestrator.services.processes import start_process
from orchestrator.utils.errors import StartPredicateError


@strawberry.type(description="Process status")
class ProcessCreated:
    id: UUID


@strawberry.input(description="Payload for process create")
class Payload:
    payload: JSON = strawberry.field(description="Payload")


async def resolve_start_process(
    info: OrchestratorInfo, name: str, payload: Payload, reporter: str | None = None
) -> ProcessCreated | MutationError:
    broadcast_func = graphql_broadcast_process_data(info)

    current_user = None
    if user_resolver := info.context.get_current_user:
        current_user = await user_resolver
    user = resolve_user_name(reporter=reporter, resolved_user=current_user)

    try:
        process_id = start_process(name, user_inputs=payload.payload, user=user, broadcast_func=broadcast_func)  # type: ignore
    except StartPredicateError as exc:
        return MutationError(message="Start predicate not satisfied", details=str(exc))
    except Exception as exc:
        return MutationError(message="Could not create process", details=str(exc))

    return ProcessCreated(id=process_id)


@strawberry.type(description="Process mutations")
class ProcessMutation:
    start_process = authenticated_mutation_field(
        resolver=resolve_start_process,
        description="Create a process in the Orchestrator",
    )



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Query building and execution module."""

from orchestrator.search.aggregations import TemporalGrouping

from . import engine
from .builder import (
    ComponentInfo,
    LeafInfo,
    build_aggregation_query,
    build_candidate_query,
    build_paths_query,
    process_path_rows,
)
from .exceptions import (
    EmptyFilterPathError,
    IncompatibleAggregationTypeError,
    IncompatibleFilterTypeError,
    IncompatibleTemporalGroupingTypeError,
    InvalidEntityPrefixError,
    InvalidLtreePatternError,
    PathNotFoundError,
    QueryValidationError,
)
from .queries import AggregateQuery, CountQuery, ExportQuery, Query, SelectQuery
from .results import (
    MatchingField,
    QueryResultsResponse,
    ResultRow,
    SearchResponse,
    SearchResult,
    VisualizationType,
    format_aggregation_response,
    format_search_response,
    generate_highlight_indices,
)
from .state import QueryState

__all__ = [
    # Builder functions
    "build_aggregation_query",
    "build_candidate_query",
    "build_paths_query",
    "process_path_rows",
    # Builder metadata
    "ComponentInfo",
    "LeafInfo",
    # Engine
    "engine",
    # Exceptions
    "EmptyFilterPathError",
    "IncompatibleAggregationTypeError",
    "IncompatibleFilterTypeError",
    "IncompatibleTemporalGroupingTypeError",
    "InvalidEntityPrefixError",
    "InvalidLtreePatternError",
    "PathNotFoundError",
    "QueryValidationError",
    # Query models
    "AggregateQuery",
    "CountQuery",
    "ExportQuery",
    "Query",
    "SelectQuery",
    "TemporalGrouping",
    # Results
    "QueryResultsResponse",
    "ResultRow",
    "MatchingField",
    "SearchResponse",
    "SearchResult",
    "VisualizationType",
    "format_aggregation_response",
    "format_search_response",
    "generate_highlight_indices",
    # State
    "QueryState",
]



from typing import Iterable

from prometheus_client import Metric
from prometheus_client.metrics_core import GaugeMetricFamily, StateSetMetricFamily
from prometheus_client.registry import Collector

from orchestrator.schemas.engine_settings import GlobalStatusEnum
from orchestrator.services import settings
from orchestrator.services.worker_status_monitor import get_worker_status_monitor


def _get_engine_status() -> tuple[GlobalStatusEnum, int]:
    """Query for getting the current status of the workflow engine.

    This includes the engine status, and the amount of currently running processes.
    """
    monitor = get_worker_status_monitor()
    running_count = monitor.get_running_jobs_count()

    engine_settings = settings.get_engine_settings_table()
    engine_status = settings.generate_engine_global_status(engine_settings, running_count)

    return engine_status, running_count


class WorkflowEngineCollector(Collector):
    """Initialize a Prometheus enum and a gauge.

    The enum of the current workflow engine status takes three values:
        - RUNNING
        - PAUSING
        - PAUSED

    This metric also exports the amount of currently running processes.
    """

    def collect(self) -> Iterable[Metric]:
        current_engine_status, running_process_count = _get_engine_status()

        engine_status = StateSetMetricFamily(
            "wfo_engine_status",
            documentation="Current workflow engine status.",
            value={status: status == current_engine_status for status in GlobalStatusEnum.values()},
        )

        engine_process_count = GaugeMetricFamily(
            "wfo_active_process_count",
            unit="count",
            value=running_process_count,
            documentation="Number of currently running processes in the workflow engine.",
        )

        return [engine_status, engine_process_count]



# Copyright 2019-2026 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from sqlalchemy import func, select

from orchestrator.db import ProcessTable, db
from orchestrator.workflow import PredicateContext, RunPredicateFail, RunPredicatePass, RunPredicateResult


def no_uncompleted_instance(context: PredicateContext) -> RunPredicateResult:
    """Predicate that prevents starting if an uncompleted instance of the workflow exists.

    Args:
        context: PredicateContext containing the workflow information.

    Returns:
        RunPredicatePass if no uncompleted instances exist, or RunPredicateFail with reason otherwise.
    """
    workflow_name = context.workflow_key
    uncompleted_count = db.session.scalar(
        select(func.count())
        .select_from(ProcessTable)
        .filter(
            ProcessTable.workflow.has(name=workflow_name),
            ProcessTable.last_status != "completed",
        )
    )
    if uncompleted_count == 0:
        return RunPredicatePass()
    return RunPredicateFail(f"Workflow '{workflow_name}' already has {uncompleted_count} uncompleted instance(s)")



import asyncio
from datetime import UTC, datetime, timedelta
from typing import Annotated
from uuid import UUID

import structlog
from pydantic import StringConstraints
from sqlalchemy import select
from typing_extensions import TypedDict

from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.db import db
from orchestrator.db.models import WorkflowTable
from orchestrator.forms.validators import Choice
from orchestrator.forms.validators.timestamp import to_timestamp_field
from orchestrator.workflow import Workflow
from orchestrator.workflows import get_workflow
from pydantic_forms.types import FormGenerator, FormGeneratorAsync, State
from pydantic_forms.validators.components.read_only import read_only_field

logger = structlog.get_logger(__name__)


class ScheduleTypeEnum(Choice):
    DATE = "Once"
    INTERVAL = "Interval"
    CRON = "Cron"


class Intervals(Choice):
    ONE_HOUR = "1 hour"
    TWO_HOURS = "2 hours"
    FOUR_HOURS = "4 hours"
    TWELVE_HOURS = "12 hours"
    TWENTY_FOUR_HOURS = "24 hours"
    ONE_WEEK = "1 week"
    TWO_WEEKS = "2 weeks"
    ONE_MONTH = "1 months"


class ButtonConfig(TypedDict, total=False):
    text: str


class Buttons(TypedDict):
    previous: ButtonConfig
    next: ButtonConfig


INTERVAL_MAPPING = {
    Intervals.ONE_HOUR: {"hours": 1},
    Intervals.TWO_HOURS: {"hours": 2},
    Intervals.FOUR_HOURS: {"hours": 4},
    Intervals.TWELVE_HOURS: {"hours": 12},
    Intervals.TWENTY_FOUR_HOURS: {"hours": 24},
    Intervals.ONE_WEEK: {"weeks": 1},
    Intervals.TWO_WEEKS: {"weeks": 2},
    Intervals.ONE_MONTH: {"weeks": 4},
}

cron_regex = r"^(?:(\*|([0-5]?\d))(?:/(\d+))?\s+){4}(?:(\*|([0-5]?\d))(?:/(\d+))?\s+)?(?:([0-9,/*\-?LW#]+)(?:\s+([0-9,/*\-?LW#]+))?(?:\s+([0-9,/*\-?LW#]+))?)$"


def get_interval_kwargs(form_data: dict) -> dict:
    return {"start_date": form_data["start_date"]} | INTERVAL_MAPPING[form_data["interval"]]


def parse_cron_field(value: str) -> int | None:
    try:
        return int(value)
    except ValueError:
        return None


def get_cron_kwargs(form_data: dict) -> dict:
    minute, hour, day, month, day_of_week = form_data["cron"].split(" ")

    return {
        "start_date": form_data["start_date"],
        "minute": parse_cron_field(minute),
        "hour": parse_cron_field(hour),
        "day": parse_cron_field(day),
        "month": parse_cron_field(month),
        "day_of_week": parse_cron_field(day_of_week),
    }


def get_tasks(user_model: OIDCUserModel | None) -> dict[str, tuple[Workflow, UUID]]:
    tasks = db.session.scalars(select(WorkflowTable))
    return {
        task_row.name: (workflow, task_row.workflow_id)
        for task_row in tasks
        if (workflow := get_workflow(task_row.name)) and workflow.authorize_callback(user_model)
    }


def has_initial_form(task: Workflow) -> bool:
    try:
        gen = task.initial_input_form({})
        form = next(gen)
        if form.model_fields:
            return True
    except StopIteration:
        return False
    except Exception as exc:
        logger.debug("Unexpected error when trying task.initial_input_form", task=task.name, error=exc)
        return True
    return False


def form_generator_send(form_generator: FormGenerator, data: dict | None) -> tuple[bool, dict]:
    try:
        return False, form_generator.send(data)  # type: ignore
    except StopIteration as e:
        return True, e.value


async def configure_schedule_form(state: State) -> FormGeneratorAsync:
    from orchestrator.forms import FormPage, SubmitFormPage

    user_model = OIDCUserModel(**_user_model) if (_user_model := state.get("user_model")) else None
    tasks = get_tasks(user_model)
    task_choices = {name: task[0].description for name, task in tasks.items()}

    class ScheduleTaskChoiceForm(FormPage):
        task: Choice("TaskChoices", task_choices)  # type: ignore

    schedule_task_form = yield ScheduleTaskChoiceForm

    task_name = schedule_task_form.task.name
    task, workflow_id = tasks[task_name]

    user_inputs = []
    if has_initial_form(task):
        form_state = {"workflow_target": task.target.value, "workflow_name": task.name} | state
        loop = asyncio.get_event_loop()
        sync_gen = task.initial_input_form(form_state)
        data = None
        while True:
            done, result = await loop.run_in_executor(None, form_generator_send, sync_gen, data)
            if done:
                form_state = result
                break
            data = yield result
            user_inputs.append(data)

    class ScheduleTypeForm(FormPage):
        task: read_only_field(schedule_task_form.task)  # type: ignore
        schedule_type: ScheduleTypeEnum

    schedule_type_form = yield ScheduleTypeForm

    current_date = datetime.now(UTC)
    default_timestamp = int((current_date + timedelta(hours=1)).timestamp())
    DateTimeField = to_timestamp_field(min_date=current_date)

    class ScheduleDateForm(SubmitFormPage):
        task: read_only_field(schedule_type_form.task)  # type: ignore
        schedule_type: read_only_field(schedule_type_form.schedule_type)  # type: ignore

        start_date: DateTimeField = default_timestamp  # type: ignore
        if schedule_type_form.schedule_type == ScheduleTypeEnum.INTERVAL:
            interval: Intervals
        if schedule_type_form.schedule_type == ScheduleTypeEnum.CRON:
            cron: Annotated[str, StringConstraints(pattern=cron_regex)]

        buttons: Buttons = {"previous": {}, "next": {"text": "Create Schedule"}}

    schedule_date_form = yield ScheduleDateForm
    _schedule_type_data = schedule_date_form.model_dump()
    schedule_type_data = _schedule_type_data | {
        "start_date": datetime.fromtimestamp(_schedule_type_data["start_date"], UTC)
    }

    match schedule_date_form.schedule_type:
        case ScheduleTypeEnum.INTERVAL:
            trigger_kwargs = get_interval_kwargs(schedule_type_data)
        case ScheduleTypeEnum.CRON:
            trigger_kwargs = get_cron_kwargs(schedule_type_data)
        case _:
            trigger_kwargs = {"run_date": schedule_type_data["start_date"]}

    yield {
        "workflow_id": workflow_id,
        "workflow_name": task_name,
        "trigger": schedule_type_data["schedule_type"].name.lower(),
        "trigger_kwargs": trigger_kwargs,
        "user_inputs": user_inputs,
        "scheduled_type": "create",
        "name": task.description,
    }



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from decimal import Decimal

import structlog
from sqlalchemy import BindParameter, Numeric, Select, literal

from orchestrator.search.core.types import EntityType, FieldType, SearchMetadata
from orchestrator.search.query.queries import ExportQuery, SelectQuery

from ..pagination import PageCursor

logger = structlog.get_logger(__name__)


class Retriever(ABC):
    """Abstract base class for applying a ranking strategy to a search query."""

    SCORE_PRECISION = 12
    SCORE_NUMERIC_TYPE = Numeric(38, 12)
    HIGHLIGHT_TEXT_LABEL = "highlight_text"
    HIGHLIGHT_PATH_LABEL = "highlight_path"
    SCORE_LABEL = "score"
    SEARCHABLE_FIELD_TYPES = [
        FieldType.STRING.value,
        FieldType.UUID.value,
        FieldType.BLOCK.value,
        FieldType.RESOURCE_TYPE.value,
    ]

    @classmethod
    def route(
        cls,
        query: "SelectQuery | ExportQuery",
        cursor: PageCursor | None,
        query_embedding: list[float] | None = None,
    ) -> "Retriever":
        """Route to the appropriate retriever instance based on query plan.

        Selects the retriever type based on available search criteria:
        - Hybrid: both embedding and fuzzy term available
        - Semantic: only embedding available
        - Fuzzy: only text term available (or fallback when embedding generation fails)
        - Structured: only filters available

        Args:
            query: SelectQuery or ExportQuery with search criteria
            cursor: Pagination cursor for cursor-based paging
            query_embedding: Query embedding for semantic search, or None if not available

        Returns:
            A concrete retriever instance based on available search criteria
        """

        from .fuzzy import FuzzyRetriever
        from .hybrid import RrfHybridRetriever
        from .process import ProcessHybridRetriever
        from .semantic import SemanticRetriever
        from .structured import StructuredRetriever

        fuzzy_term = query.fuzzy_term
        is_process = query.entity_type == EntityType.PROCESS

        # If vector_query exists but embedding generation failed, fall back to fuzzy search with full query text
        if query_embedding is None and query.vector_query is not None and query.query_text is not None:
            fuzzy_term = query.query_text

        # Select retriever based on available search criteria
        if query_embedding is not None and fuzzy_term is not None:
            if is_process:
                return ProcessHybridRetriever(query_embedding, fuzzy_term, cursor)
            return RrfHybridRetriever(query_embedding, fuzzy_term, cursor)
        if query_embedding is not None:
            return SemanticRetriever(query_embedding, cursor)
        if fuzzy_term is not None:
            if is_process:
                return ProcessHybridRetriever(None, fuzzy_term, cursor)
            return FuzzyRetriever(fuzzy_term, cursor)

        return StructuredRetriever(cursor, query.order_by)

    @abstractmethod
    def apply(self, candidate_query: Select) -> Select:
        """Apply the ranking logic to the given candidate query.

        Args:
            candidate_query (Select): A SQLAlchemy `Select` statement returning candidate entity IDs.

        Returns:
            Select: A new `Select` statement with ranking expressions applied.
        """
        ...

    def _quantize_score_for_pagination(self, score_value: float) -> BindParameter[Decimal]:
        """Convert score value to properly quantized Decimal parameter for pagination."""
        quantizer = Decimal(1).scaleb(-self.SCORE_PRECISION)
        pas_dec = Decimal(str(score_value)).quantize(quantizer)
        return literal(pas_dec, type_=self.SCORE_NUMERIC_TYPE)

    @property
    @abstractmethod
    def metadata(self) -> SearchMetadata:
        """Return metadata describing this search strategy."""
        ...



from datetime import datetime
from typing import Annotated, Callable

import structlog
from pydantic.functional_validators import AfterValidator

from pydantic_forms.validators import timestamp

logger = structlog.get_logger(__name__)


DATE_FORMAT = "DD-MMM-YYYY HH:mm z"  # 19-Mar-2026 00:00 UTC
TIME_FORMAT = "HH:mm"  # 00:00


def validator_min_date(min_date: datetime | None) -> Callable[[int], int]:
    def _validate_min_date(timestamp: int) -> int:
        if min_date and timestamp < min_date.timestamp():
            raise ValueError("start_date must not be in the past")
        return timestamp

    return _validate_min_date


def to_timestamp_field(min_date: datetime | None = None) -> type:
    min_value = int(min_date.timestamp()) if min_date else None
    timestamp_field = timestamp(
        locale="nl-nl",
        min=min_value,
        validate=False,  # We want to raise custom error messages
        date_format=DATE_FORMAT,
        time_format=TIME_FORMAT,
    )

    return Annotated[
        timestamp_field,  # type: ignore
        AfterValidator(validator_min_date(min_date)),
    ]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import secrets
import string
from pathlib import Path
from typing import Literal

from pydantic import Field, NonNegativeInt, PostgresDsn, RedisDsn, Secret, SecretStr
from pydantic.main import BaseModel
from pydantic_settings import BaseSettings

from oauth2_lib.fastapi import OIDCUserModel
from oauth2_lib.settings import oauth2lib_settings
from orchestrator.services.settings_env_variables import expose_settings
from orchestrator.utils.auth import Authorizer
from pydantic_forms.types import strEnum

SecretRedisDsn = Secret[RedisDsn]
SecretPostgresDsn = Secret[PostgresDsn]


class ExecutorType(strEnum):
    WORKER = "celery"
    THREADPOOL = "threadpool"


class LifecycleValidationMode(strEnum):
    STRICT = "strict"
    LOOSE = "loose"
    IGNORED = "ignored"


class AppSettings(BaseSettings):
    TESTING: bool = True
    SESSION_SECRET: SecretStr = "".join(secrets.choice(string.ascii_letters) for i in range(16))  # type: ignore
    CORS_ORIGINS: str = "*"
    CORS_ALLOW_METHODS: list[str] = ["GET", "PUT", "PATCH", "POST", "DELETE", "OPTIONS", "HEAD"]
    CORS_ALLOW_HEADERS: list[str] = ["If-None-Match", "Authorization", "If-Match", "Content-Type"]
    CORS_EXPOSE_HEADERS: list[str] = [
        "Cache-Control",
        "Content-Language",
        "Content-Length",
        "Content-Type",
        "Expires",
        "Last-Modified",
        "Pragma",
        "Content-Range",
        "ETag",
    ]
    ENVIRONMENT: str = "local"
    EXECUTOR: str = ExecutorType.THREADPOOL
    WORKFLOWS_SWAGGER_HOST: str = "localhost"
    WORKFLOWS_GUI_URI: str = "http://localhost:3000"
    BASE_URL: str = "http://localhost:8080"  # Base URL for the API (used for generating export URLs)
    DATABASE_URI: SecretPostgresDsn = "postgresql://nwa:nwa@localhost/orchestrator-core"  # type: ignore
    MAX_WORKERS: int = 5
    WORKER_STATUS_INTERVAL: int = Field(
        5, description="Interval in seconds for updating worker status count in the background monitor"
    )
    MAIL_SERVER: str = "localhost"
    MAIL_PORT: int = 25
    MAIL_STARTTLS: bool = False
    CACHE_URI: SecretRedisDsn = "redis://localhost:6379/0"  # type: ignore
    REDIS_RETRY_COUNT: NonNegativeInt = Field(
        2, description="Number of retries for redis connection errors/timeouts, 0 to disable"
    )  # More info: https://redis-py.readthedocs.io/en/stable/retry.html
    ENABLE_DISTLOCK_MANAGER: bool = True
    DISTLOCK_BACKEND: str = "memory"
    CC_NOC: int = 0
    SERVICE_NAME: str = "orchestrator-core"
    LOGGING_HOST: str = "localhost"
    LOG_LEVEL: str = "DEBUG"
    SLACK_ENGINE_SETTINGS_HOOK_ENABLED: bool = False
    SLACK_ENGINE_SETTINGS_HOOK_URL: str = ""
    TRACING_ENABLED: bool = False
    TRACE_HOST: str = "http://localhost:4317"
    TRANSLATIONS_DIR: Path | None = None
    WEBSOCKET_BROADCASTER_URL: SecretStr = "memory://"  # type: ignore
    ENABLE_WEBSOCKETS: bool = True
    DISABLE_INSYNC_CHECK: bool = False
    DEFAULT_PRODUCT_WORKFLOWS: list[str] = ["modify_note"]
    SKIP_MODEL_FOR_MIGRATION_DB_DIFF: list[str] = []
    SERVE_GRAPHQL_UI: str | None = "graphiql"
    FEDERATION_VERSION: Literal[
        "2.0", "2.1", "2.2", "2.3", "2.4", "2.5", "2.6", "2.7", "2.8", "2.9", "2.10", "2.11"
    ] = "2.9"
    DEFAULT_CUSTOMER_FULLNAME: str = "Default::Orchestrator-Core Customer"
    DEFAULT_CUSTOMER_SHORTCODE: str = "default-cust"
    DEFAULT_CUSTOMER_IDENTIFIER: str = "59289a57-70fb-4ff5-9c93-10fe67b12434"
    TASK_LOG_RETENTION_DAYS: int = 3
    ENABLE_GRAPHQL_DEPRECATION_CHECKER: bool = True
    ENABLE_GRAPHQL_PROFILING_EXTENSION: bool = False
    ENABLE_GRAPHQL_STATS_EXTENSION: bool = False
    ENABLE_PROMETHEUS_METRICS_ENDPOINT: bool = False
    VALIDATE_OUT_OF_SYNC_SUBSCRIPTIONS: bool = False
    FILTER_BY_MODE: Literal["partial", "exact"] = "exact"
    EXPOSE_SETTINGS: bool = False
    EXPOSE_OAUTH_SETTINGS: bool = False
    LIFECYCLE_VALIDATION_MODE: LifecycleValidationMode = LifecycleValidationMode.LOOSE


app_settings = AppSettings()

# Set oauth2lib_settings variables to the same (default) value of settings
oauth2lib_settings.SERVICE_NAME = app_settings.SERVICE_NAME
oauth2lib_settings.ENVIRONMENT = app_settings.ENVIRONMENT

if app_settings.EXPOSE_SETTINGS:
    expose_settings("app_settings", app_settings)
if app_settings.EXPOSE_OAUTH_SETTINGS:
    expose_settings("oauth2lib_settings", oauth2lib_settings)


class Authorizers(BaseModel):
    # Callbacks specifically for orchestrator-core callbacks.
    # Separate from defaults for user-defined workflows and steps.
    internal_authorize_callback: Authorizer | None = None
    internal_retry_auth_callback: Authorizer | None = None

    async def authorize_callback(self, user: OIDCUserModel | None) -> bool:
        """This is the authorize_callback to be registered for workflows defined within orchestrator-core.

        If Authorizers.internal_authorize_callback is None, this function will return True.
        i.e. any user will be authorized to start internal workflows.
        """
        if self.internal_authorize_callback is None:
            return True
        return await self.internal_authorize_callback(user)

    async def retry_auth_callback(self, user: OIDCUserModel | None) -> bool:
        """This is the retry_auth_callback to be registered for workflows defined within orchestrator-core.

        If Authorizers.internal_retry_auth_callback is None, this function will return True.
        i.e. any user will be authorized to retry internal workflows on failure.
        """
        if self.internal_retry_auth_callback is None:
            return True
        return await self.internal_retry_auth_callback(user)


_authorizers = Authorizers()


def get_authorizers() -> Authorizers:
    """Acquire singleton of app authorizers to assign these callbacks at app setup.

    Ensures downstream users can acquire singleton without being tempted to do
    from orchestrator.settings import authorizers
    authorizers = my_authorizers
    or
    from orchestrator import settings
    settings.authorizers = my_authorizers

    ...each of which goes wrong in its own way.
    """
    return _authorizers



import strawberry

from orchestrator.schemas import SubscriptionDescriptionSchema


@strawberry.experimental.pydantic.type(model=SubscriptionDescriptionSchema, all_fields=True)
class CustomerDescription:
    pass



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from orchestrator.search.agent.state import SearchState

import structlog
from pydantic import BaseModel, ConfigDict
from pydantic_ai.messages import ModelRequest, ModelResponse, SystemPromptPart, TextPart, UserPromptPart

logger = structlog.get_logger(__name__)


class MemoryScope(Enum):
    """Defines what context is visible to different steps."""

    FULL = "full"  # User questions + answers + full execution traces (Planner)
    LIGHTWEIGHT = "lightweight"  # User questions + answers + full query JSON (Search/Aggregation)
    MINIMAL = "minimal"  # User questions + answers + query_id only (Result Actions)


@dataclass
class Step:
    """Base class for execution steps in a turn.

    A step represents a discrete action taken during agent execution,
    either visiting a node or calling a tool.
    """

    step_type: str  # Specific type: step name or tool name
    description: str
    success: bool = True
    error_message: str | None = None
    timestamp: datetime = field(default_factory=lambda: datetime.now())


@dataclass
class AgentStep(Step):
    """Records an agent step execution.

    Contains nested tool_steps that were executed within this step.
    """

    tool_steps: list["ToolStep"] = field(default_factory=list)  # Tools called by this agent.

    def add_tool_step(self, tool_step: "ToolStep") -> None:
        """Add a tool step that was executed within this step."""
        self.tool_steps.append(tool_step)


@dataclass
class ToolStep(Step):
    """Records a tool call execution."""

    context: dict[str, Any] | None = None


@dataclass
class Turn:
    user_question: str
    assistant_answer: str | None = None
    steps: list[AgentStep] = field(default_factory=list)
    current_step: AgentStep | None = None
    timestamp: datetime = field(default_factory=lambda: datetime.now())

    @property
    def is_complete(self) -> bool:
        return self.assistant_answer is not None


class Memory(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    turns: list[Turn] = []

    @property
    def current_turn(self) -> Turn | None:
        """Return the last turn if it's incomplete, else None."""
        if self.turns and not self.turns[-1].is_complete:
            return self.turns[-1]
        return None

    @property
    def completed_turns(self) -> list[Turn]:
        """Return all completed turns."""
        return [t for t in self.turns if t.is_complete]

    def start_turn(self, user_question: str) -> None:
        self.turns.append(Turn(user_question=user_question))

    def start_step(self, step_name: str) -> None:
        """Start a new agent step (finishes previous step if any)."""
        if not self.current_turn:
            raise ValueError("No active turn")

        if self.current_turn.current_step:
            self.finish_step()

        self.current_turn.current_step = AgentStep(
            step_type=step_name,
            description=f"Executing {step_name}",
        )

    def record_tool_step(self, tool_step: ToolStep) -> None:
        """Record a tool call within the current step."""
        if not self.current_turn:
            raise ValueError("No active turn")
        if not self.current_turn.current_step:
            raise ValueError("No active step — must call start_step first")

        self.current_turn.current_step.add_tool_step(tool_step)

    def finish_step(self) -> None:
        """Finish the current step and add it to the list."""
        if not self.current_turn:
            raise ValueError("No active turn")
        if not self.current_turn.current_step:
            raise ValueError("No active step to finish")

        turn = self.current_turn
        if turn.current_step is not None:
            turn.steps.append(turn.current_step)
            turn.current_step = None

    def complete_turn(self, assistant_answer: str) -> None:
        if not self.current_turn:
            raise ValueError("No active turn")

        # Finish any in-progress step
        if self.current_turn.current_step:
            self.finish_step()

        turn = self.current_turn

        logger.debug(
            "complete_turn",
            completed_count_before=len(self.completed_turns),
            question=turn.user_question,
            step_count=len(turn.steps),
        )

        turn.assistant_answer = assistant_answer

    def _format_query_summary(self, steps: list[AgentStep], include_full_query: bool = True) -> str | None:
        """Format query summary - either full JSON or just query_id with one-liner.

        Args:
            steps: List of AgentStep objects to extract queries from
            include_full_query: If True, show full query JSON; if False, just query_id + description

        Returns:
            Formatted query summary string, or None if no queries
        """
        import json

        # Find tool steps with query_ids
        query_tools = []
        for step in steps:
            for tool_step in step.tool_steps:
                if tool_step.context and tool_step.context.get("query_id"):
                    query_tools.append(tool_step)

        if not query_tools:
            return None

        summaries = []
        for tool in query_tools:
            ctx = tool.context or {}
            query_id = ctx.get("query_id")
            query_snapshot = ctx.get("query_snapshot")
            if include_full_query and query_snapshot:
                # Full query JSON for Search/Aggregation nodes
                query_json = json.dumps(query_snapshot, indent=2)
                summaries.append(f"Query {query_id}:\n{query_json}")
            else:
                # Minimal one-liner for ResultActions
                summaries.append(f"Query {query_id}: {tool.description}")

        return "\n\n".join(summaries) if summaries else None

    def _format_execution_trace(self, steps: list[AgentStep]) -> str | None:
        """Format steps as execution trace string.

        Args:
            steps: List of AgentStep objects to format

        Returns:
            Formatted execution trace string, or None if no action steps
        """
        # Filter out Planner - only show action steps
        action_steps = [step for step in steps if step.step_type != "Planner"]

        if not action_steps:
            return None

        lines = ["Plan executed:"]
        for i, step in enumerate(action_steps, 1):
            lines.append(f"  {i}. [Step] {step.step_type}")
            if step.tool_steps:
                for tool_step in step.tool_steps:
                    query_id = tool_step.context.get("query_id") if tool_step.context else None
                    query_info = f" [query: {query_id}]" if query_id else ""
                    lines.append(f"     • [Tool] {tool_step.step_type}: {tool_step.description}{query_info}")

        return "\n".join(lines)

    def _format_turn_context(self, steps: list[AgentStep], scope: MemoryScope) -> str | None:
        """Format turn context based on memory scope."""
        if not steps:
            return None
        if scope == MemoryScope.FULL:
            return self._format_execution_trace(steps)
        if scope == MemoryScope.LIGHTWEIGHT:
            return self._format_query_summary(steps, include_full_query=True)
        # MINIMAL: just query_id + one-liner
        return self._format_query_summary(steps, include_full_query=False)

    def get_message_history(
        self, max_turns: int = 5, scope: MemoryScope = MemoryScope.FULL
    ) -> list[ModelRequest | ModelResponse]:
        """Get conversation history as pydantic-ai message objects for message_history parameter.

        Args:
            max_turns: Maximum number of recent turns to include
            scope: Memory scope controlling what details to include

        Returns:
            List of ModelRequest/ModelResponse messages for agent message_history
        """
        recent = self.completed_turns[-max_turns:]

        # Build messages using pydantic-ai message types
        messages: list[ModelRequest | ModelResponse] = []
        for turn in recent:
            # User message
            messages.append(ModelRequest(parts=[UserPromptPart(content=turn.user_question)]))

            # Add context based on scope
            context = self._format_turn_context(turn.steps, scope)
            if context:
                messages.append(ModelRequest(parts=[SystemPromptPart(content=context)]))

            # Assistant response
            if turn.assistant_answer is not None:
                messages.append(ModelResponse(parts=[TextPart(content=turn.assistant_answer)]))

        # Always add current turn's user question
        if self.current_turn:
            messages.append(ModelRequest(parts=[UserPromptPart(content=self.current_turn.user_question)]))

        return messages

    def format_current_steps(self) -> str:
        """Format the steps taken in the current turn for display in prompts.

        Used to show what has been executed so far in this incomplete turn.
        """
        if not self.current_turn:
            return "None"

        all_steps = list(self.current_turn.steps)
        if self.current_turn.current_step:
            all_steps.append(self.current_turn.current_step)

        trace = self._format_execution_trace(all_steps)
        return trace if trace else "None"

    def format_context_for_llm(
        self,
        state: SearchState,
        *,
        include_current_run_steps: bool = False,
    ) -> str:
        """Context formatter for LLM prompts.

        Current user request and conversation history are passed via
        message_history parameter, not included here.

        Args:
            state: SearchState instance for accessing current context
            include_current_run_steps: Show steps executed so far in current run

        Returns:
            Formatted context string ready to insert into prompt
        """
        if include_current_run_steps:
            steps = self.format_current_steps()
            return f"# Steps Already Executed\n{steps}"

        return ""



# Copyright 2019-2025 SURF, ESnet, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer

from orchestrator.cli import (
    database,
    generate,
    scheduler,
)
from orchestrator.llm_settings import llm_settings

app = typer.Typer()
app.add_typer(scheduler.app, name="scheduler", help="Access all the scheduler functions")
app.add_typer(database.app, name="db", help="Interact with the application database")
app.add_typer(generate.app, name="generate", help="Generate products, workflows and other artifacts")

if llm_settings.SEARCH_ENABLED:
    from orchestrator.cli.search import index_llm, resize_embedding, search_explore, speedtest

    app.add_typer(index_llm.app, name="index", help="(Re-)Index the search table.")
    app.add_typer(search_explore.app, name="search", help="Try out different search types.")
    app.add_typer(
        resize_embedding.app,
        name="embedding",
        help="Resize the vector dimension of the embedding column in the search table.",
    )
    app.add_typer(
        speedtest.app,
        name="speedtest",
        help="Search performance testing and analysis.",
    )


if __name__ == "__main__":
    app()



"""set is_task of task_validate_product_type to true.

Revision ID: aa1b92d0d07d
Revises: 961eddbd4c13
Create Date: 2026-03-10 14:03:36.981457

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "aa1b92d0d07d"
down_revision = "d69e10434a04"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()
    conn.execute(sa.text("UPDATE workflows SET is_task=true WHERE name = 'task_validate_product_type'"))


def downgrade() -> None:
    pass



import asyncio

import structlog
import typer
from pydantic import ValidationError

from orchestrator.cli.search.display import display_filtered_paths_only, display_results
from orchestrator.db import db
from orchestrator.search.core.types import EntityType, FilterOp, UIType
from orchestrator.search.filters import EqualityFilter, FilterTree, LtreeFilter, PathFilter
from orchestrator.search.query import engine
from orchestrator.search.query.queries import SelectQuery
from orchestrator.search.query.validation import get_structured_filter_schema

app = typer.Typer(help="Experiment with the subscription search indexes.")

logger = structlog.getLogger(__name__)


@app.command()
def structured(path: str, value: str, entity_type: EntityType = EntityType.SUBSCRIPTION, limit: int = 10) -> None:
    """Finds subscriptions where a specific field path contains an exact value.

    Example:
        dotenv run python main.py search structured "subscription.status" "provisioning"
        ...
        {
            "path": "subscription.status",
            "value": "provisioning"
        },
        ...
    """
    path_filter = PathFilter(path=path, condition=EqualityFilter(op=FilterOp.EQ, value=value), value_kind=UIType.STRING)
    query = SelectQuery(entity_type=entity_type, filters=FilterTree.from_flat_and([path_filter]), limit=limit)
    search_response = asyncio.run(engine.execute_search(query=query, db_session=db.session))
    display_filtered_paths_only(search_response.results, query, db.session)
    display_results(search_response.results, db.session, "Match")


@app.command()
def semantic(query_text: str, entity_type: EntityType = EntityType.SUBSCRIPTION, limit: int = 10) -> None:
    """Finds subscriptions that are conceptually most similar to the query text.

    Example:
        dotenv run python main.py search semantic "Shop for an alligator store"
        ...
        {
            "path": "subscription.shop.shop_description",
            "value": "Kingswood reptiles shop"
        },
        ...
    """
    query = SelectQuery(entity_type=entity_type, query_text=query_text, limit=limit)
    search_response = asyncio.run(engine.execute_search(query=query, db_session=db.session))
    display_results(search_response.results, db.session, "Distance")


@app.command()
def fuzzy(term: str, entity_type: EntityType = EntityType.SUBSCRIPTION, limit: int = 10) -> None:
    """Finds subscriptions containing text similar to the query, tolerating typos.

    Example:
        dotenv run python main.py search fuzzy "Colonel"
        ...
        {
          "path": "description",
          "value": "X Follower WF for TimCoronel"
        },
        ...
    """
    query = SelectQuery(entity_type=entity_type, query_text=term, limit=limit)
    search_response = asyncio.run(engine.execute_search(query=query, db_session=db.session))
    display_results(search_response.results, db.session, "Similarity")


@app.command()
def hierarchical(
    op: str = typer.Argument(..., help="The hierarchical operation to perform."),
    path: str = typer.Argument(..., help="The ltree path or lquery pattern for the operation."),
    query_text: str | None = typer.Option(None, "--query", "-q", help="An optional fuzzy term to rank the results."),
    entity_type: EntityType = EntityType.SUBSCRIPTION,
    limit: int = 10,
) -> None:
    """Performs a hierarchical search, optionally combined with fuzzy ranking.

    Examples:
        dotenv run python main.py search hierarchical is_descendant "subscription.shop" --query "Kingwood"
        dotenv run python main.py search hierarchical matches_lquery "*.x_follower.x_follower_status*"
    """
    try:
        condition = LtreeFilter(value=path, op=op)  # type: ignore[arg-type]
    except (ValueError, ValidationError) as e:
        raise typer.BadParameter(f"Invalid filter: {e}")

    path_filter = PathFilter(path="ltree_hierarchical_filter", condition=condition, value_kind=UIType.STRING)

    query = SelectQuery(
        entity_type=entity_type, filters=FilterTree.from_flat_and([path_filter]), query_text=query_text, limit=limit
    )
    search_response = asyncio.run(engine.execute_search(query=query, db_session=db.session))
    display_results(search_response.results, db.session, "Hierarchical Score")


@app.command()
def hybrid(query_text: str, term: str, entity_type: EntityType = EntityType.SUBSCRIPTION, limit: int = 10) -> None:
    """Performs a hybrid search, combining semantic and fuzzy matching.

    Example:
        dotenv run python main.py search hybrid "reptile store" "Kingswood"
    """
    query = SelectQuery(entity_type=entity_type, query_text=query_text, limit=limit)
    logger.info("Executing Hybrid Search", query_text=query_text, term=term)
    search_response = asyncio.run(engine.execute_search(query=query, db_session=db.session))
    display_results(search_response.results, db.session, "Hybrid Score")


@app.command("generate-schema")
def generate_schema() -> None:
    """Generates and prints the dynamic filter schema from the live search index.

    This queries the index for all distinct non-string paths to be used as
    context for the LLM agent.

    Example:
        dotenv run python main.py search generate-schema
    """

    schema_map = get_structured_filter_schema()

    if not schema_map:
        logger.warning("No filterable paths found in the search index.")
        return

    logger.info("\nAvailable Structured Filters:\n")
    for path, value_type in schema_map.items():
        logger.info(f"- {path}: {value_type}")

    logger.info("Successfully generated dynamic schema.", path_count=len(schema_map))


@app.command("nested-demo")
def nested_demo(entity_type: EntityType = EntityType.SUBSCRIPTION, limit: int = 10) -> None:
    tree = FilterTree.model_validate(
        {
            "op": "AND",
            "children": [
                {
                    "op": "OR",
                    "children": [
                        # First OR case: Active subscriptions from 2024
                        {
                            "op": "AND",
                            "children": [
                                {
                                    "path": "subscription.status",
                                    "condition": {"op": "eq", "value": "active"},
                                    "value_kind": "string",
                                },
                                {
                                    "path": "subscription.start_date",
                                    "condition": {
                                        "op": "between",
                                        "value": {
                                            "start": "2024-01-01T00:00:00Z",
                                            "end": "2024-12-31T23:59:59Z",
                                        },
                                    },
                                    "value_kind": "datetime",
                                },
                            ],
                        },
                        # Second OR case: Terminated subscriptions before 2026
                        {
                            "op": "AND",
                            "children": [
                                {
                                    "path": "subscription.status",
                                    "condition": {"op": "eq", "value": "terminated"},
                                    "value_kind": "string",
                                },
                                {
                                    "path": "subscription.end_date",
                                    "condition": {"op": "lte", "value": "2025-12-31"},
                                    "value_kind": "datetime",
                                },
                            ],
                        },
                    ],
                },
                {
                    "path": "subscription.*.port_mode",
                    "condition": {"op": "matches_lquery", "value": "*.port_mode"},
                    "value_kind": "string",
                },
            ],
        }
    )

    query = SelectQuery(entity_type=entity_type, filters=tree, limit=limit)
    search_response = asyncio.run(engine.execute_search(query=query, db_session=db.session))

    display_results(search_response.results, db.session, "Score")


if __name__ == "__main__":
    app()



# Copyright 2019-2026 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from importlib import import_module

import structlog

from orchestrator.utils.docs import make_workflow_index_doc
from orchestrator.workflow import Workflow

logger = structlog.get_logger(__name__)

DEFAULT_PKG = "orchestrator.workflows"


ALL_WORKFLOWS: dict[str, "LazyWorkflowInstance"] = {}


class LazyWorkflowInstance:
    """Capture workflow function reference for lazy instantiation.

    Importing all workflow functions into a higher level package was the cause for many circular import problems. To
    remedy this we register each workflow with an instance of this class and only import them when required.

    Importing only is not sufficient as some workflow definitions are created indirectly by helper functions that
    might or might not be parameterized. Hence the usage of the phrase 'instantiation' (which is not entirely
    technically correct, but as my imagination is limited and naming is one of the hard problem in programming,
    that's what it is for now).

    Usage::

        LazyWorkflowInstance(<package ref>, <name>, <is_callable>)

    Both `<package ref>` and `<name>` are strings. `<package ref>` can be an absolute package reference or a
    relative package reference. In case of the latter, the package it is relative to is indicated by the module level
    constant `DEFAULT_PKG`.

    Examples::

         LazyWorkflowInstance(".port", "create")  # from .port import create

    The `".port"` is short for `"server.workflows.port"`. If a different parent package is required, use the
    absolute reference syntax::

        LazyWorkflowInstance("different.package.port", "create")

    """

    package: str
    function: str

    def __init__(self, package: str, name: str) -> None:
        self.package = package
        self.function = name
        ALL_WORKFLOWS[name] = self

    def instantiate(self) -> Workflow:
        """Import and instantiate a workflow and return it.

        This can be as simple as merely importing a workflow function. However, if it concerns a workflow generating
        function, that function will be called with or without arguments as specified.

        Returns:
            A workflow function.

        """
        try:
            if self.package.startswith("."):
                # relative import, hence `package` should be set
                mod = import_module(self.package, package=DEFAULT_PKG)
            else:
                mod = import_module(self.package)
        except ModuleNotFoundError:
            module_name = f"{DEFAULT_PKG}{self.package}" if self.package.startswith(".") else self.package
            raise ValueError(f"Invalid workflow: module {module_name} does not exist or has invalid imports")

        wf = getattr(mod, self.function)

        # Set workflow name here to make them complete
        wf.name = self.function
        return wf

    def __str__(self) -> str:
        return f"{self.package}.{self.function}"

    def __repr__(self) -> str:
        return f"LazyWorkflowInstance('{self.package}','{self.function}')"


def get_workflow(name: str) -> Workflow | None:
    wi = ALL_WORKFLOWS.get(name)

    if not wi:
        logger.debug("LazyWorkflowInstance not found with this name", name=name)
        return None

    return wi.instantiate()


# Modify
LazyWorkflowInstance(".modify_note", "modify_note")

# Tasks
LazyWorkflowInstance(".tasks.cleanup_tasks_log", "task_clean_up_tasks")
LazyWorkflowInstance(".tasks.resume_workflows", "task_resume_workflows")
LazyWorkflowInstance(".tasks.validate_products", "task_validate_products")
LazyWorkflowInstance(".tasks.validate_product_type", "task_validate_product_type")
LazyWorkflowInstance(".tasks.validate_subscriptions", "task_validate_subscriptions")

__doc__ = make_workflow_index_doc(ALL_WORKFLOWS)



from uuid import UUID

from sqlalchemy import select
from sqlalchemy.orm import Session
from structlog import get_logger

from orchestrator.db.models import GraphSnapshotTable
from orchestrator.search.agent.state import SearchState

logger = get_logger(__name__)


class PostgresStatePersistence:
    """PostgreSQL state persistence for search agent.

    Stores state snapshots in the graph_snapshots table, allowing execution to be
    interrupted and resumed across conversation turns.
    Uses thread_id for persistence so state continues across multiple runs in the same thread.
    """

    def __init__(self, thread_id: str, run_id: UUID, session: Session):
        self.thread_id = thread_id
        self.run_id = run_id
        self.session = session
        self._sequence_counter = 0

    async def snapshot(self, state: SearchState) -> None:
        """Save a state snapshot to PostgreSQL.

        Args:
            state: Current agent state
        """
        snapshot_data = {
            "state": state.model_dump(mode="json"),
            "sequence": self._sequence_counter,
        }

        db_snapshot = GraphSnapshotTable(
            run_id=self.run_id,
            sequence_number=self._sequence_counter,
            snapshot_data=snapshot_data,
        )

        self.session.add(db_snapshot)
        self.session.flush()

        logger.debug(
            "Saved state snapshot",
            run_id=str(self.run_id),
            sequence=self._sequence_counter,
        )

        self._sequence_counter += 1

    async def load_state(self) -> SearchState | None:
        """Load the most recent state for this thread.

        Returns:
            The deserialized SearchState, or None if no snapshots exist
        """
        from orchestrator.db.models import AgentRunTable

        stmt = (
            select(GraphSnapshotTable)
            .join(AgentRunTable, GraphSnapshotTable.run_id == AgentRunTable.run_id)
            .where(AgentRunTable.thread_id == self.thread_id)
            .order_by(AgentRunTable.created_at.desc(), GraphSnapshotTable.sequence_number.desc())
            .limit(1)
        )

        result = self.session.execute(stmt)
        db_snapshot = result.scalar_one_or_none()

        if not db_snapshot:
            logger.debug("No snapshots found for thread", thread_id=self.thread_id)
            return None

        snapshot_data = db_snapshot.snapshot_data
        state_data = snapshot_data.get("state", {})
        state = SearchState.model_validate(state_data)

        # Update sequence counter to continue from where we left off
        self._sequence_counter = db_snapshot.sequence_number + 1

        logger.debug(
            "Loaded state for resume",
            thread_id=self.thread_id,
            run_id=str(self.run_id),
            sequence=db_snapshot.sequence_number,
        )

        return state



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
from collections.abc import Generator, Iterable, Iterator
from contextlib import contextmanager, nullcontext
from functools import lru_cache
from typing import Any

import structlog
from litellm.utils import encode, get_max_tokens
from sqlalchemy import delete, tuple_
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.dialects.postgresql.dml import Insert
from sqlalchemy.orm import Session
from sqlalchemy_utils.types.ltree import Ltree

from orchestrator.db import db
from orchestrator.db.models import AiSearchIndex
from orchestrator.llm_settings import llm_settings
from orchestrator.search.core.embedding import EmbeddingIndexer
from orchestrator.search.core.types import ExtractedField, IndexableRecord
from orchestrator.search.indexing.registry import EntityConfig
from orchestrator.search.indexing.traverse import DatabaseEntity

logger = structlog.get_logger(__name__)


@contextmanager
def _maybe_begin(session: Session | None) -> Iterator[None]:
    if session is None:
        yield
    else:
        with session.begin():
            yield


@contextmanager
def _maybe_progress(show_progress: bool, total_count: int | None, label: str) -> Iterator[Any]:
    """Context manager that optionally creates a progress bar."""
    if show_progress:
        import typer

        with typer.progressbar(
            length=total_count,
            label=label,
            show_eta=True,
            show_percent=bool(total_count),
        ) as progress:
            yield progress
    else:
        yield None


class Indexer:
    """Index entities into `AiSearchIndex` using streaming reads and batched writes.

    Entities are read from a streaming iterator and accumulated into chunks of
    size `chunk_size`. For each chunk, the indexer extracts fields, diffs via
    content hashes, deletes stale paths, and prepares upserts using a two-list
    buffer:
    - Embeddable list (STRING fields): maintains a running token count against a
        token budget (model context window minus a safety margin) and flushes when
        adding the next item would exceed the budget.
    - Non-embeddable list: accumulated in parallel and does not contribute to the
        flush condition.
    Each flush (or end-of-chunk) emits a single combined UPSERT batch from both
    lists (wrapped in a per-chunk transaction in non-dry-runs).

    Args:
        config (EntityConfig): Registry config describing the entity kind,
            ORM table, and traverser.
        dry_run (bool): If True, skip DELETE/UPSERT statements and external
            embedding calls.
        force_index (bool): If True, ignore existing hashes and reindex all
            fields for each entity.
        chunk_size (int): Number of entities to process per batch. Defaults to 1000.

    Notes:
        - Non-dry-run runs open a write session and wrap each processed chunk in
          a transaction (`Session.begin()`).
        - Read queries use the passed session when available, otherwise the
          generic `db.session`.

    Workflow:
        1) Stream entities (yield_per=chunk_size) and accumulate into a chunk.
        2) Begin transaction for the chunk.
        3) determine_changes() → fields_to_upsert, paths_to_delete.
        4) Delete stale paths.
        5) Build UPSERT batches with a two-list buffer:
        - Embeddable list (STRING): track running token count; flush when next item
            would exceed the token budget (model max context - safety margin).
        - Non-embeddable list: accumulate in parallel; does not affect flushing.
        6) Execute UPSERT for each batch (skip in dry_run).
        7) Commit transaction (auto on context exit).
        8) Repeat until the stream is exhausted.
    """

    def __init__(
        self,
        config: EntityConfig,
        dry_run: bool,
        force_index: bool,
        chunk_size: int = 1000,
        show_progress: bool = False,
        total_count: int | None = None,
    ) -> None:
        self.config = config
        self.dry_run = dry_run
        self.force_index = force_index
        self.chunk_size = chunk_size
        self.show_progress = show_progress
        self.total_count = total_count
        self.embedding_model = llm_settings.EMBEDDING_MODEL
        self.logger = logger.bind(entity_kind=config.entity_kind.value)
        self._entity_titles: dict[str, str] = {}

    def run(self, entities: Iterable[DatabaseEntity]) -> int:
        """Orchestrates the entire indexing process."""
        chunk: list[DatabaseEntity] = []
        total_records_processed = 0
        total_identical_records = 0

        write_scope = db.database_scope() if not self.dry_run else nullcontext()

        def flush() -> None:
            nonlocal total_records_processed, total_identical_records
            with _maybe_begin(session):
                processed_in_chunk, identical_in_chunk = self._process_chunk(chunk, session)
                total_records_processed += processed_in_chunk
                total_identical_records += identical_in_chunk
            chunk.clear()

        with write_scope as database:
            session: Session | None = getattr(database, "session", None)

            with _maybe_progress(
                self.show_progress, self.total_count, f"Indexing {self.config.entity_kind.value}"
            ) as progress:
                for entity in entities:
                    chunk.append(entity)

                    if len(chunk) >= self.chunk_size:
                        flush()
                        if progress:
                            progress.update(self.chunk_size)

                if chunk:
                    flush()
                    if progress:
                        progress.update(len(chunk))

        final_log_message = (
            f"processed {total_records_processed} records and skipped {total_identical_records} identical records."
        )
        self.logger.info(
            f"Dry run, would have indexed {final_log_message}"
            if self.dry_run
            else f"Indexing done, {final_log_message}"
        )
        return total_records_processed

    def _process_chunk(self, entity_chunk: list[DatabaseEntity], session: Session | None = None) -> tuple[int, int]:
        """Process a chunk of entities."""
        if not entity_chunk:
            return 0, 0

        self._entity_titles.clear()

        fields_to_upsert, paths_to_delete, identical_count = self._determine_changes(entity_chunk, session)

        if paths_to_delete and session is not None:
            self.logger.debug(f"Deleting {len(paths_to_delete)} stale records in chunk.")
            self._execute_batched_deletes(paths_to_delete, session)

        if fields_to_upsert:
            upsert_stmt = self._get_upsert_statement()
            batch_generator = self._generate_upsert_batches(fields_to_upsert)

            for batch in batch_generator:
                if self.dry_run:
                    self.logger.debug(f"Dry Run: Would upsert {len(batch)} records.")
                elif batch and session:
                    session.execute(upsert_stmt, batch)

        return len(fields_to_upsert), identical_count

    def _determine_changes(
        self, entities: list[DatabaseEntity], session: Session | None = None
    ) -> tuple[list[tuple[str, ExtractedField]], list[tuple[str, Ltree]], int]:
        """Identifies all changes across all entities using pre-fetched data."""
        entity_ids = [str(getattr(e, self.config.pk_name)) for e in entities]
        read_session = session or db.session
        existing_hashes = {} if self.force_index else self._get_all_existing_hashes(entity_ids, read_session)

        fields_to_upsert: list[tuple[str, ExtractedField]] = []
        paths_to_delete: list[tuple[str, Ltree]] = []
        identical_records_count = 0

        for entity in entities:
            entity_id = str(getattr(entity, self.config.pk_name))
            current_fields = self.config.traverser.get_fields(
                entity, pk_name=self.config.pk_name, root_name=self.config.root_name
            )

            entity_title = self.config.get_title_from_fields(current_fields)
            self._entity_titles[entity_id] = entity_title

            entity_hashes = existing_hashes.get(entity_id, {})
            current_paths = set()

            for field in current_fields:
                current_paths.add(field.path)
                current_hash = self._compute_content_hash(field.path, field.value, field.value_type, entity_title)
                if field.path not in entity_hashes or entity_hashes[field.path] != current_hash:
                    fields_to_upsert.append((entity_id, field))
                else:
                    identical_records_count += 1

            stale_paths = set(entity_hashes.keys()) - current_paths
            paths_to_delete.extend([(entity_id, Ltree(p)) for p in stale_paths])

        return fields_to_upsert, paths_to_delete, identical_records_count

    def _execute_batched_deletes(self, paths_to_delete: list[tuple[str, Ltree]], session: Session) -> None:
        """Execute delete operations in batches to avoid PostgreSQL stack depth limits."""
        for i in range(0, len(paths_to_delete), self.chunk_size):
            batch = paths_to_delete[i : i + self.chunk_size]
            delete_stmt = delete(AiSearchIndex).where(tuple_(AiSearchIndex.entity_id, AiSearchIndex.path).in_(batch))
            session.execute(delete_stmt)
            self.logger.debug(f"Deleted batch of {len(batch)} records.")

    def _get_all_existing_hashes(self, entity_ids: list[str], session: Session) -> dict[str, dict[str, str]]:
        """Fetches all existing hashes for a list of entity IDs in a single query."""
        if not entity_ids:
            return {}

        results = (
            session.query(AiSearchIndex.entity_id, AiSearchIndex.path, AiSearchIndex.content_hash)
            .filter(AiSearchIndex.entity_id.in_(entity_ids))
            .all()
        )

        hashes_by_entity: dict[str, dict[str, str]] = {eid: {} for eid in entity_ids}
        for entity_id, path, content_hash in results:
            hashes_by_entity[str(entity_id)][str(path)] = content_hash
        return hashes_by_entity

    def _generate_upsert_batches(
        self, fields_to_upsert: Iterable[tuple[str, ExtractedField]]
    ) -> Generator[list[IndexableRecord], None, None]:
        """Streams through fields, buffers them by token count, and yields batches."""
        embeddable_buffer: list[tuple[str, ExtractedField]] = []
        non_embeddable_records: list[IndexableRecord] = []
        current_tokens = 0

        max_ctx = self._get_max_tokens()
        safe_margin = int(max_ctx * llm_settings.EMBEDDING_SAFE_MARGIN_PERCENT)
        token_budget = max(1, max_ctx - safe_margin)

        max_batch_size = llm_settings.EMBEDDING_MAX_BATCH_SIZE

        for entity_id, field in fields_to_upsert:
            if field.value_type.is_embeddable(field.value):
                text = self._prepare_text_for_embedding(field)
                try:
                    item_tokens = len(encode(model=self.embedding_model, text=text))
                except Exception as e:
                    self.logger.warning("Tokenization failed; skipping.", path=field.path, err=str(e))
                    continue

                if item_tokens > max_ctx:
                    self.logger.warning(
                        "Field exceeds context; skipping.", path=field.path, tokens=item_tokens, max_ctx=max_ctx
                    )
                    continue

                should_flush = embeddable_buffer and (
                    current_tokens + item_tokens > token_budget
                    or (max_batch_size and len(embeddable_buffer) >= max_batch_size)
                )

                if should_flush:
                    yield self._flush_buffer(embeddable_buffer, non_embeddable_records)
                    embeddable_buffer.clear()
                    non_embeddable_records.clear()
                    current_tokens = 0

                embeddable_buffer.append((entity_id, field))
                current_tokens += item_tokens
            else:
                record = self._make_indexable_record(field, entity_id, embedding=None)
                non_embeddable_records.append(record)

        if embeddable_buffer or non_embeddable_records:
            yield self._flush_buffer(embeddable_buffer, non_embeddable_records)

    def _flush_buffer(self, embeddable_buffer: list, non_embeddable_records: list) -> list[IndexableRecord]:
        """Processes and combines buffers into a single batch."""
        if not embeddable_buffer:
            return non_embeddable_records

        texts_to_embed = [self._prepare_text_for_embedding(f) for _, f in embeddable_buffer]
        embeddings = EmbeddingIndexer.get_embeddings_from_api_batch(texts_to_embed, self.dry_run)

        if len(embeddable_buffer) != len(embeddings):
            raise ValueError(f"Embedding mismatch: sent {len(embeddable_buffer)}, received {len(embeddings)}")

        with_embeddings = [
            self._make_indexable_record(field, entity_id, embedding)
            for (entity_id, field), embedding in zip(embeddable_buffer, embeddings)
        ]
        return non_embeddable_records + with_embeddings

    def _get_max_tokens(self) -> int:
        """Gets max tokens, using a fallback from settings if necessary."""
        try:
            max_ctx = get_max_tokens(self.embedding_model)
            if isinstance(max_ctx, int):
                return max_ctx
        except Exception:
            # Allow local(unknown) models to fall back.
            self.logger.warning("Could not auto-detect max tokens.", model=self.embedding_model)

        max_ctx = llm_settings.EMBEDDING_FALLBACK_MAX_TOKENS
        if not isinstance(max_ctx, int):
            raise RuntimeError("Model not recognized and EMBEDDING_FALLBACK_MAX_TOKENS not set.")
        self.logger.warning("Using configured fallback token limit.", fallback=max_ctx)
        return max_ctx

    @staticmethod
    def _prepare_text_for_embedding(field: ExtractedField) -> str:
        return f"{field.path}: {str(field.value)}"

    @staticmethod
    def _compute_content_hash(path: str, value: Any, value_type: Any, entity_title: str = "") -> str:
        v = "" if value is None else str(value)
        content = f"{path}:{v}:{value_type}:{entity_title}"
        return hashlib.sha256(content.encode("utf-8")).hexdigest()

    def _make_indexable_record(
        self, field: ExtractedField, entity_id: str, embedding: list[float] | None
    ) -> IndexableRecord:
        entity_title = self._entity_titles[entity_id]
        return IndexableRecord(
            entity_id=entity_id,
            entity_type=self.config.entity_kind.value,
            entity_title=entity_title,
            path=Ltree(field.path),
            value=field.value,
            value_type=field.value_type,
            content_hash=self._compute_content_hash(field.path, field.value, field.value_type, entity_title),
            embedding=embedding if embedding else None,
        )

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_upsert_statement() -> Insert:
        stmt = insert(AiSearchIndex)
        return stmt.on_conflict_do_update(
            index_elements=[AiSearchIndex.entity_id, AiSearchIndex.path],
            set_={
                AiSearchIndex.entity_title: stmt.excluded.entity_title,
                AiSearchIndex.value: stmt.excluded.value,
                AiSearchIndex.value_type: stmt.excluded.value_type,
                AiSearchIndex.content_hash: stmt.excluded.content_hash,
                AiSearchIndex.embedding: stmt.excluded.embedding,
            },
        )
